import os
from pathlib import Path
from textwrap import dedent
from unittest import TestCase
from unittest.mock import AsyncMock

import pytest

from mentat.code_context import CodeContext
from mentat.config import Config
from mentat.errors import ReturnToUser
from mentat.feature_filters.default_filter import DefaultFilter
from mentat.git_handler import get_non_gitignored_files
from mentat.include_files import is_file_text_encoded
from mentat.interval import Interval
from mentat.llm_api_handler import count_tokens
from tests.conftest import run_git_command


@pytest.mark.asyncio
async def test_path_gitignoring(temp_testbed, mock_code_context):
    gitignore_path = ".gitignore"
    testing_dir_path = "git_testing_dir"
    os.makedirs(testing_dir_path)

    # create 3 files, 2 ignored in gitignore, 1 not
    ignored_file_path_1 = Path(os.path.join(testing_dir_path, "ignored_file_1.txt"))
    ignored_file_path_2 = Path(os.path.join(testing_dir_path, "ignored_file_2.txt"))
    non_ignored_file_path = Path(os.path.join(testing_dir_path, "non_ignored_file.txt"))

    with open(gitignore_path, "w") as gitignore_file:
        gitignore_file.write("ignored_file_1.txt\nignored_file_2.txt")

    for file_path in [ignored_file_path_1, ignored_file_path_2, non_ignored_file_path]:
        with open(file_path, "w") as file:
            file.write("I am a file")

    # Run CodeFileManager on the git_testing_dir, and also explicitly pass in ignored_file_2.txt
    mock_code_context.include(testing_dir_path)
    mock_code_context.include(ignored_file_path_2)

    expected_file_paths = [
        os.path.join(temp_testbed, ignored_file_path_2),
        os.path.join(temp_testbed, non_ignored_file_path),
    ]

    case = TestCase()
    file_paths = [
        str(file_path.resolve()) for file_path in mock_code_context.include_files
    ]
    case.assertListEqual(sorted(expected_file_paths), sorted(file_paths))


@pytest.mark.asyncio
async def test_bracket_file(temp_testbed, mock_code_context):
    file_path_1 = Path("[file].tsx")
    file_path_2 = Path("test:[file].tsx")

    with file_path_1.open("w") as file_1:
        file_1.write("Testing")
    with file_path_2.open("w") as file_2:
        file_2.write("Testing")

    mock_code_context.include(file_path_1)
    mock_code_context.include(file_path_2)
    expected_file_paths = [
        temp_testbed / file_path_1,
        temp_testbed / file_path_2,
    ]

    case = TestCase()
    file_paths = list(mock_code_context.include_files.keys())
    case.assertListEqual(sorted(expected_file_paths), sorted(file_paths))


@pytest.mark.asyncio
async def test_config_glob_exclude(mocker, temp_testbed, mock_code_context):
    # Makes sure glob exclude config works
    mocker.patch.object(
        Config, "file_exclude_glob_list", new=[os.path.join("glob_test", "**", "*.py")]
    )

    glob_exclude_path = os.path.join("glob_test", "bagel", "apple", "exclude_me.py")
    glob_include_path = os.path.join("glob_test", "bagel", "apple", "include_me.ts")
    directly_added_glob_excluded_path = Path(
        os.path.join("glob_test", "bagel", "apple", "directly_added_glob_excluded.py")
    )
    os.makedirs(os.path.dirname(glob_exclude_path), exist_ok=True)
    with open(glob_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am excluded")
    with open(glob_include_path, "w") as glob_include_file:
        glob_include_file.write("I am included")
    with open(
        directly_added_glob_excluded_path, "w"
    ) as directly_added_glob_excluded_file:
        directly_added_glob_excluded_file.write(
            "Config excludes me but I'm included if added directly"
        )

    mock_code_context.include(".")
    mock_code_context.include(directly_added_glob_excluded_path)

    assert Path(temp_testbed / glob_exclude_path) not in mock_code_context.include_files
    assert Path(temp_testbed / glob_include_path) in mock_code_context.include_files
    assert (
        Path(temp_testbed / directly_added_glob_excluded_path)
        in mock_code_context.include_files
    )


@pytest.mark.asyncio
async def test_glob_include(temp_testbed, mock_code_context):
    # Make sure glob include works
    glob_include_path = os.path.join("glob_test", "bagel", "apple", "include_me.py")
    glob_include_path2 = os.path.join("glob_test", "bagel", "apple", "include_me2.py")
    glob_exclude_path = os.path.join("glob_test", "bagel", "apple", "exclude_me.ts")

    os.makedirs(os.path.dirname(glob_include_path), exist_ok=True)
    with open(glob_include_path, "w") as glob_include_file:
        glob_include_file.write("I am included")
    os.makedirs(os.path.dirname(glob_include_path2), exist_ok=True)
    with open(glob_include_path2, "w") as glob_include_file:
        glob_include_file.write("I am also included")
    os.makedirs(os.path.dirname(glob_exclude_path), exist_ok=True)
    with open(glob_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am not included")

    mock_code_context.include("**/*.py")

    file_paths = [
        str(file_path.resolve()) for file_path in mock_code_context.include_files
    ]
    assert os.path.join(temp_testbed, glob_exclude_path) not in file_paths
    assert os.path.join(temp_testbed, glob_include_path) in file_paths
    assert os.path.join(temp_testbed, glob_include_path2) in file_paths


@pytest.mark.asyncio
async def test_cli_glob_exclude(temp_testbed, mock_code_context):
    # Make sure cli glob exclude works and overrides regular include
    glob_include_then_exclude_path = os.path.join(
        "glob_test", "bagel", "apple", "include_then_exclude_me.py"
    )
    glob_exclude_path = os.path.join("glob_test", "bagel", "apple", "exclude_me.ts")

    os.makedirs(os.path.dirname(glob_include_then_exclude_path), exist_ok=True)
    with open(glob_include_then_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am included then excluded")
    os.makedirs(os.path.dirname(glob_exclude_path), exist_ok=True)
    with open(glob_exclude_path, "w") as glob_exclude_file:
        glob_exclude_file.write("I am excluded")

    mock_code_context.include("**/*.py", exclude_patterns=["**/*.py", "**/*.ts"])

    file_paths = [file_path for file_path in mock_code_context.include_files]
    assert os.path.join(temp_testbed, glob_include_then_exclude_path) not in file_paths
    assert os.path.join(temp_testbed, glob_exclude_path) not in file_paths


@pytest.mark.asyncio
async def test_text_encoding_checking(temp_testbed, mock_session_context):
    # Makes sure we don't include non text encoded files, and we quit if user gives us one
    nontext_path = "iamnottext.py"
    with open(nontext_path, "wb") as f:
        # 0x81 is invalid in UTF-8 (single byte > 127), and undefined in cp1252 and iso-8859-1
        f.write(bytearray([0x81]))

    code_context = CodeContext(
        mock_session_context.stream,
        mock_session_context.cwd,
    )
    code_context.include("./")
    file_paths = [file_path for file_path in code_context.include_files]
    assert os.path.join(temp_testbed, nontext_path) not in file_paths

    nontext_path_requested = "iamalsonottext.py"
    with open(nontext_path_requested, "wb") as f:
        # 0x81 is invalid in UTF-8 (single byte > 127), and undefined in cp1252 and iso-8859-1
        f.write(bytearray([0x81]))
    code_context = CodeContext(
        mock_session_context.stream,
        mock_session_context.cwd,
    )
    code_context.include(nontext_path_requested)
    assert not code_context.include_files


@pytest.mark.asyncio
@pytest.mark.clear_testbed
async def test_max_auto_tokens(mocker, temp_testbed, mock_session_context):
    with open("file_1.py", "w") as f:
        f.write(dedent("""\
            def func_1(x, y):
                return x + y
            
            def func_2():
                return 3
            """))

    with open("file_2.py", "w") as f:
        f.write(dedent("""\
            def func_3(a, b, c):
                return a * b ** c
            
            def func_4(string):
                print(string)
            """))
    run_git_command(temp_testbed, "add", ".")
    run_git_command(temp_testbed, "commit", "-m", "initial commit")

    code_context = CodeContext(
        mock_session_context.stream,
        mock_session_context.code_context.git_root,
    )
    code_context.include("file_1.py")
    mock_session_context.config.auto_context_tokens = 8000
    filter_mock = AsyncMock(side_effect=lambda features: features)
    mocker.patch.object(DefaultFilter, "filter", side_effect=filter_mock)

    async def _count_max_tokens_where(tokens_used: int) -> int:
        code_message = await code_context.get_code_message(tokens_used, prompt="prompt")
        return count_tokens(code_message, "gpt-4", full_message=True)

    assert await _count_max_tokens_where(0) == 89  # Code
    with pytest.raises(ReturnToUser):
        await _count_max_tokens_where(1e6)


@pytest.mark.clear_testbed
def test_get_all_features(temp_testbed, mock_code_context):
    # Create a sample file
    path1 = Path(temp_testbed) / "sample_path1.py"
    path2 = Path(temp_testbed) / "sample_path2.py"
    with open(path1, "w") as file1:
        file1.write("def sample_function():\n    pass\n")
    with open(path2, "w") as file2:
        file2.write("def sample_function():\n    pass\n")

    # Test without include_files
    features = mock_code_context.get_all_features()
    assert len(features) == 2
    feature1 = next(f for f in features if f.path == path1)
    feature2 = next(f for f in features if f.path == path2)
    for _f, _p in zip((feature1, feature2), (path1, path2)):
        feature = next(f for f in features if f.path == _p)
        assert feature.path == _p

    # Test with include_files argument matching one file
    mock_code_context.include(path1)
    features = mock_code_context.get_all_features(split_intervals=False)
    assert len(features) == 2
    feature1b = next(f for f in features if f.path == path1)
    feature2b = next(f for f in features if f.path == path2)
    assert feature1b.interval.whole_file()
    assert feature2b.interval.whole_file()


@pytest.mark.asyncio
async def test_get_code_message_ignore(mocker, temp_testbed, mock_session_context):
    mock_session_context.config.auto_context_tokens = 8000
    mocker.patch.object(Config, "maximum_context", new=7000)
    filter_mock = AsyncMock(side_effect=lambda features: features)
    mocker.patch.object(DefaultFilter, "filter", side_effect=filter_mock)
    code_context = CodeContext(
        mock_session_context.stream,
        temp_testbed,
        ignore_patterns=["scripts", "**/*.txt"],
    )
    code_message = await code_context.get_code_message(0, prompt="prompt")

    # Iterate through all files in temp_testbed; if they're not in the ignore
    # list, they should be in the code message.
    for file in get_non_gitignored_files(temp_testbed):
        abs_path = temp_testbed / file
        rel_path = abs_path.relative_to(temp_testbed).as_posix()
        if (
            not is_file_text_encoded(abs_path)
            or "scripts" in rel_path
            or rel_path.endswith(".txt")
        ):
            assert rel_path not in code_message
        else:
            assert rel_path in code_message


@pytest.mark.no_git_testbed
def test_include_single_file_interval(mock_code_context):
    mock_code_context.include("multifile_calculator/calculator.py:10-12")

    assert len(mock_code_context.include_files) == 1
    multifile_calculator_path = Path("multifile_calculator/calculator.py").resolve()
    assert multifile_calculator_path in mock_code_context.include_files
    assert len(mock_code_context.include_files[multifile_calculator_path]) == 1
    assert mock_code_context.include_files[multifile_calculator_path][
        0
    ].interval == Interval(10, 12)


@pytest.mark.no_git_testbed
def test_include_multiple_file_intervals(mock_code_context):
    mock_code_context.include("multifile_calculator/calculator.py:10-12")
    mock_code_context.include("multifile_calculator/calculator.py:14-20")

    assert len(mock_code_context.include_files) == 1
    multifile_calculator_path = Path("multifile_calculator/calculator.py").resolve()
    assert len(mock_code_context.include_files[multifile_calculator_path]) == 2
    assert mock_code_context.include_files[multifile_calculator_path][
        0
    ].interval == Interval(10, 12)
    assert mock_code_context.include_files[multifile_calculator_path][
        1
    ].interval == Interval(14, 20)


@pytest.mark.no_git_testbed
def test_include_missing_file_interval(mock_code_context):
    mock_code_context.include("multifile_calculator/calculator.py:9000-9001")
    assert len(mock_code_context.include_files) == 0


@pytest.mark.no_git_testbed
def test_include_overlapping_file_intervals(mock_code_context):
    mock_code_context.include("multifile_calculator/calculator.py:1-5")
    mock_code_context.include("multifile_calculator/calculator.py:1-6")
    assert len(mock_code_context.include_files) == 1
    multifile_calculator_path = Path("multifile_calculator/calculator.py").resolve()
    assert len(mock_code_context.include_files[multifile_calculator_path]) == 2
    assert mock_code_context.include_files[multifile_calculator_path][
        0
    ].interval == Interval(1, 5)
    assert mock_code_context.include_files[multifile_calculator_path][
        1
    ].interval == Interval(1, 6)


@pytest.mark.no_git_testbed
def test_include_duplicate_file_interval(mock_code_context):
    mock_code_context.include("multifile_calculator/calculator.py:1-5")
    mock_code_context.include("multifile_calculator/calculator.py:1-5")
    assert len(mock_code_context.include_files) == 1
    multifile_calculator_path = Path("multifile_calculator/calculator.py").resolve()
    assert len(mock_code_context.include_files[multifile_calculator_path]) == 1
    assert mock_code_context.include_files[multifile_calculator_path][
        0
    ].interval == Interval(1, 5)


@pytest.mark.no_git_testbed
def test_exclude_single_file_interval(mock_code_context):
    mock_code_context.include("multifile_calculator/calculator.py:10-12")
    mock_code_context.exclude("multifile_calculator/calculator.py:10-12")
    assert len(mock_code_context.include_files) == 0


@pytest.mark.no_git_testbed
def test_exclude_multiple_file_intervals(mock_code_context):
    mock_code_context.include("multifile_calculator/calculator.py:1-5")
    mock_code_context.include("multifile_calculator/calculator.py:6-10")
    mock_code_context.exclude("multifile_calculator/calculator.py:1-5")
    assert len(mock_code_context.include_files) == 1
    multifile_calculator_path = Path("multifile_calculator/calculator.py").resolve()
    assert len(mock_code_context.include_files[multifile_calculator_path]) == 1
    assert mock_code_context.include_files[multifile_calculator_path][
        0
    ].interval == Interval(6, 10)


@pytest.mark.no_git_testbed
def test_exclude_missing_file_interval(mock_code_context):
    mock_code_context.include("multifile_calculator/calculator.py:1-5")
    mock_code_context.exclude("multifile_calculator/calculator.py:3-10")
    assert len(mock_code_context.include_files) == 1
    multifile_calculator_path = Path("multifile_calculator/calculator.py").resolve()
    assert len(mock_code_context.include_files[multifile_calculator_path]) == 1
    assert mock_code_context.include_files[multifile_calculator_path][
        0
    ].interval == Interval(1, 5)


@pytest.mark.no_git_testbed
def test_include_single_directory(mock_code_context):
    mock_code_context.include(
        "multifile_calculator", exclude_patterns=["**/__pycache__"]
    )
    assert len(mock_code_context.include_files) == 3
    assert (
        Path("multifile_calculator/__init__.py").resolve()
        in mock_code_context.include_files
    )
    assert (
        Path("multifile_calculator/calculator.py").resolve()
        in mock_code_context.include_files
    )
    assert (
        Path("multifile_calculator/operations.py").resolve()
        in mock_code_context.include_files
    )


@pytest.mark.no_git_testbed
def test_include_duplicate_directory(mock_code_context):
    mock_code_context.include(
        "multifile_calculator", exclude_patterns=["**/__pycache__"]
    )
    mock_code_context.include(
        "multifile_calculator", exclude_patterns=["**/__pycache__"]
    )
    assert len(mock_code_context.include_files) == 3


@pytest.mark.no_git_testbed
def test_include_missing_directory(mock_code_context):
    mock_code_context.include(
        "multifile_calculator", exclude_patterns=["**/__pycache__"]
    )
    mock_code_context.include("this_directory_does_not_exist")
    assert len(mock_code_context.include_files) == 3


@pytest.mark.no_git_testbed
def test_exclude_single_directory(mock_code_context):
    mock_code_context.include(
        "multifile_calculator", exclude_patterns=["**/__pycache__"]
    )
    mock_code_context.exclude("multifile_calculator")
    assert len(mock_code_context.include_files) == 0


@pytest.mark.no_git_testbed
def test_exclude_missing_directory(mock_code_context):
    mock_code_context.exclude("this_directory_does_not_exist")
    assert len(mock_code_context.include_files) == 0



import asyncio
import logging
import shlex
from typing import Any, Coroutine

from mentat.command.command import Command
from mentat.errors import RemoteKeyboardInterrupt, SessionExit
from mentat.session_context import SESSION_CONTEXT
from mentat.session_stream import StreamMessage


async def _get_input_request(**kwargs: Any) -> StreamMessage:
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream

    message = stream.send("", channel="input_request", **kwargs)
    response = await stream.recv(f"input_request:{message.id}")
    logging.debug(f"User Input: {response.data}")
    return response


async def collect_user_input(plain: bool = False) -> StreamMessage:
    """
    Listens for user input on a new channel

    send a message requesting user to send a response
    create a new broadcast channel that listens for the input
    close the channel after receiving the input
    """

    response = await _get_input_request(plain=plain)
    # Quit on q
    if isinstance(response.data, str) and response.data.strip() == "q":
        raise SessionExit

    return response


async def ask_yes_no(default_yes: bool) -> bool:
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream

    while True:
        # TODO: combine this into a single message (include content)
        stream.send("(Y/n)" if default_yes else "(y/N)")
        response = await collect_user_input(plain=True)
        content = response.data.strip().lower()
        if content in ["y", "n", ""]:
            break
    return content == "y" or (content != "n" and default_yes)


async def collect_input_with_commands() -> StreamMessage:
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream

    response = await collect_user_input()
    while isinstance(response.data, str) and response.data.startswith("/"):
        try:
            # We only use shlex to split the arguments, not the command itself
            arguments = shlex.split(" ".join(response.data.split(" ")[1:]))
            command = Command.create_command(response.data[1:].split(" ")[0])
            await command.apply(*arguments)
        except ValueError as e:
            stream.send(f"Error processing command arguments: {e}", style="error")
        response = await collect_user_input()
    return response


async def listen_for_interrupt(
    coro: Coroutine[None, None, Any], raise_exception_on_interrupt: bool = True
):
    """Listens for an 'interrupt' message from a client

    This function is used to cancel long-running coroutines from a remote client. If a
    message is received on the "interrupt" channel for the current session stream, the
    asyncio.Task created from `coro` will be canceled.

    TODO:
      - make sure task cancellation actually cancels the tasks
      - Is there any kind of delay, or a possiblity of one?
      - make sure there's no race conditions

    The `.result()` call for `wrapped_task` will re-raise any exceptions thrown
    inside of that Task.
    """
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream

    async with stream.interrupt_lock:
        interrupt_task = asyncio.create_task(stream.recv("interrupt"))
        wrapped_task = asyncio.create_task(coro)

        done, pending = await asyncio.wait(
            {interrupt_task, wrapped_task}, return_when=asyncio.FIRST_COMPLETED
        )

        for task in pending:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

        if wrapped_task in done:
            return wrapped_task.result()
        else:
            # Send a newline for terminal clients (remove later)
            stream.send("\n")

            if raise_exception_on_interrupt:
                raise RemoteKeyboardInterrupt



from pathlib import Path

import pytest

from mentat.errors import PathValidationError
from mentat.include_files import (
    is_interval_path,
    validate_and_format_path,
    validate_file_interval_path,
    validate_file_path,
    validate_glob_path,
)


def test_is_interval_path():
    assert not is_interval_path(Path("file.py"))
    assert not is_interval_path(Path("file.py:"))
    assert is_interval_path("file.py:0-1")


def test_validate_file_path():
    # raise on non-absolute file path
    with pytest.raises(PathValidationError):
        validate_file_path(Path("file.py"))

    # raise on non-existent file path
    with pytest.raises(PathValidationError):
        validate_file_path(Path("file.py").resolve())


def test_validate_file_interval_path():
    # raise on non-absolute file interval path
    with pytest.raises(PathValidationError):
        validate_file_interval_path(Path("file.py:0-1"))

    # raise on non-existent file interval path
    with pytest.raises(PathValidationError):
        validate_file_interval_path(Path("file.py:0-1").resolve())

    # raise on no intervals
    file_path = Path("file2.txt").resolve()
    with open(file_path, "w") as file:
        file.write("hello world\nhi hi hi")
    with pytest.raises(PathValidationError):
        validate_file_interval_path(Path(f"{file_path}:"))

    # raise on invalid interval
    with pytest.raises(PathValidationError):
        validate_file_interval_path(Path(f"{file_path}:0-1,10-11"))


def test_validate_glob_path():
    # raise on non-absolute glob path
    with pytest.raises(PathValidationError):
        validate_glob_path(Path("*.py"))

    # raise on no files matching glob
    with pytest.raises(PathValidationError):
        validate_glob_path(Path("*.txt").resolve())


def test_validate_and_format_tilde_home_directory():
    with pytest.raises(PathValidationError) as e_info:
        validate_and_format_path("~/non_existing_file.py", "/fake/cwd")
    assert str(Path.home() / "non_existing_file.py") in str(e_info.value)



import asyncio
from asyncio import Event
from asyncio.tasks import Task
from pathlib import Path
from typing import List

from mentat.config import Config
from mentat.errors import MentatError
from mentat.session import Session
from mentat.session_stream import StreamMessageSource


class PythonClient:
    def __init__(
        self,
        cwd: Path = Path.cwd(),
        paths: List[Path] = [],
        exclude_paths: List[Path] = [],
        ignore_paths: List[Path] = [],
        diff: str | None = None,
        pr_diff: str | None = None,
        config: Config = Config(),
    ):
        self.cwd = cwd.expanduser().resolve()
        self.paths = paths
        self.exclude_paths = exclude_paths
        self.ignore_paths = ignore_paths
        self.diff = diff
        self.pr_diff = pr_diff
        self.config = config

        self._accumulated_message = ""
        self.stopped = Event()
        self._call_mentat_task = None

    async def _call_mentat(self, message: str):
        input_request_message = await self.session.stream.recv("input_request")
        self.session.stream.send(
            message,
            source=StreamMessageSource.CLIENT,
            channel=f"input_request:{input_request_message.id}",
        )
        if message.strip().lower() == "y":
            await self.wait_for_edit_completion()

        temp = self._accumulated_message
        self._accumulated_message = ""
        return temp

    async def call_mentat(self, message: str):
        self._call_mentat_task = asyncio.create_task(self._call_mentat(message))
        try:
            return await self._call_mentat_task
        except asyncio.CancelledError:
            print(self.session.error)
            raise MentatError("Session failed")

    async def call_mentat_auto_accept(self, message: str):
        await self.call_mentat(message)
        await self.call_mentat("y")

    async def wait_for_edit_completion(self):
        await self.session.stream.recv(channel="edits_complete")

    async def _accumulate_messages(self):
        async for message in self.session.stream.listen():
            end = "\n"
            if isinstance(message.extra.get("end"), str):
                end = message.extra["end"]
            self._accumulated_message += message.data + end

    async def _listen_for_client_exit(self):
        await self.session.stream.recv("client_exit")
        await self._stop()

    async def startup(self):
        self.session = Session(
            self.cwd,
            self.paths,
            self.exclude_paths,
            self.ignore_paths,
            self.diff,
            self.pr_diff,
            self.config,
        )
        self.session.start()
        self.acc_task = asyncio.create_task(self._accumulate_messages())
        self.exit_task: Task[None] = asyncio.create_task(self._listen_for_client_exit())

    async def shutdown(self):
        """Sends the stop signal to the session and returns when client is fully shutdown."""
        self.session.stream.send(None, channel="session_exit")
        await self.stopped.wait()

    async def _stop(self):
        self.acc_task.cancel()
        self.exit_task.cancel()
        if self._call_mentat_task:
            self._call_mentat_task.cancel()
        self.stopped.set()

    def get_conversation(self):
        return self.session.ctx.conversation

    def get_cost_tracker(self):
        return self.session.ctx.cost_tracker



from datetime import datetime
from uuid import uuid4

from mentat.session_stream import StreamMessage, StreamMessageSource
from mentat.terminal.loading import LoadingHandler


class MockTqdm:
    total = 100
    n = 0

    def __init__(self, desc):
        self.desc = desc

    def update(self, n):
        self.n += n

    def set_description(self, desc):
        self.desc = desc

    closed = False

    def close(self):
        self.closed = True


def test_loading_handler(mocker):
    mock_tqdm = MockTqdm("")
    tqdm_mock = mocker.patch("mentat.terminal.loading.tqdm")
    tqdm_mock.return_value = mock_tqdm
    lh = LoadingHandler()

    msg = StreamMessage(
        uuid4(),
        "loading",
        StreamMessageSource.SERVER,
        "Test message",
        {"progress": 50},
        datetime.utcnow(),
    )
    lh.update(msg)
    assert mock_tqdm.n == 50
    assert mock_tqdm.closed is False

    lh.update(msg)
    assert mock_tqdm.n == 100
    assert mock_tqdm.desc == "Test message"
    assert mock_tqdm.closed is True



from unittest.mock import MagicMock, patch

import pytest

from mentat.vision.vision_manager import ScreenshotException, VisionManager


@pytest.fixture
def mock_webdriver():
    with patch("selenium.webdriver.Safari") as mock:
        yield mock


@pytest.fixture
def mock_platform():
    with patch("platform.system") as mock:
        yield mock


def test_vision_manager_screenshot(mock_platform, mock_webdriver, temp_testbed):
    mock_driver_instance = MagicMock()
    mock_webdriver.return_value = mock_driver_instance

    mock_platform.return_value = "Darwin"
    mock_driver_instance.get_screenshot_as_png.return_value = b"fake_screenshot_data"

    vision_manager = VisionManager()
    vision_manager._open_browser()

    # Test taking a screenshot of an opened page
    vision_manager.screenshot()
    mock_driver_instance.get_screenshot_as_png.assert_called_once()

    # Test taking a screenshot of a specific URL
    test_url = "http://example.com"
    vision_manager.screenshot(test_url)
    mock_driver_instance.get.assert_called_with(test_url)
    assert mock_driver_instance.get_screenshot_as_png.call_count == 2

    # Test taking a screenshot of a local file
    test_file_path = "scripts/calculator.py"
    vision_manager.screenshot(test_file_path)
    mock_driver_instance.get.assert_called_with(
        f"file://{temp_testbed / test_file_path}"
    )
    assert mock_driver_instance.get_screenshot_as_png.call_count == 3

    # Test exception when no browser is open
    vision_manager.driver = None
    with pytest.raises(ScreenshotException):
        vision_manager.screenshot()



import gc
import os
import shutil
import stat
import subprocess
import tempfile
import time
from datetime import datetime
from pathlib import Path
from unittest.mock import AsyncMock
from uuid import uuid4

import pytest
from openai.types.chat import ChatCompletion, ChatCompletionChunk, ChatCompletionMessage
from openai.types.chat.chat_completion import Choice
from openai.types.chat.chat_completion_chunk import Choice as AsyncChoice
from openai.types.chat.chat_completion_chunk import ChoiceDelta

from mentat import config
from mentat.agent_handler import AgentHandler
from mentat.auto_completer import AutoCompleter
from mentat.code_context import CodeContext
from mentat.code_file_manager import CodeFileManager
from mentat.config import Config, config_file_name
from mentat.conversation import Conversation
from mentat.cost_tracker import CostTracker
from mentat.git_handler import get_git_root_for_path
from mentat.llm_api_handler import LlmApiHandler
from mentat.sampler.sampler import Sampler
from mentat.session_context import SESSION_CONTEXT, SessionContext
from mentat.session_stream import SessionStream, StreamMessage, StreamMessageSource
from mentat.streaming_printer import StreamingPrinter
from mentat.vision.vision_manager import VisionManager

pytest_plugins = ("pytest_reportlog",)


def filter_mark(items, mark, exists):
    new_items = []
    for item in items:
        marker = item.get_closest_marker(mark)
        if bool(marker) == bool(exists):
            new_items.append(item)
    return new_items


def pytest_addoption(parser):
    parser.addoption("--benchmark", action="store_true")
    parser.addoption("--uitest", action="store_true")
    # The following flags are used by benchmark tests
    parser.addoption(
        "--max_benchmarks",
        action="store",
        default="1",
        help="The maximum number of exercises to run",
    )
    parser.addoption(
        "--retries",
        action="store",
        default="1",
        help="Number of times to retry a benchmark",
    )
    parser.addoption(
        "--max_iterations",
        action="store",
        default="1",
        help="Number of times to rerun mentat with error messages",
    )
    parser.addoption(
        "--language",
        action="store",
        default="python",
        help="Which exercism language to do exercises for",
    )
    parser.addoption(
        "--max_workers",
        action="store",
        default="1",
        help="Number of workers to use for multiprocessing",
    )
    parser.addoption(
        "--refresh_repo",
        action="store_true",
        default=False,
        help="When set local changes will be discarded.",
    )
    parser.addoption(
        "--benchmarks",
        action="append",
        nargs="*",
        default=[],
        help=(
            "Which benchmarks to run. max_benchmarks ignored when set. Exact meaning"
            " depends on benchmark."
        ),
    )
    parser.addoption(
        "--repo",
        action="store",
        default="mentat",
        help="For benchmarks that are evaluated against a repo",
    )
    parser.addoption(
        "--evaluate_baseline",
        action="store_true",
        help="Evaluate the baseline for the benchmark",
    )


@pytest.fixture
def refresh_repo(request):
    return request.config.getoption("--refresh_repo")


@pytest.fixture
def benchmarks(request):
    benchmarks = request.config.getoption("--benchmarks")
    if len(benchmarks) == 1:
        return benchmarks[0]
    return benchmarks


@pytest.fixture
def max_benchmarks(request):
    return int(request.config.getoption("--max_benchmarks"))


def pytest_configure(config):
    config.addinivalue_line("markers", "benchmark: run benchmarks that call openai")
    config.addinivalue_line(
        "markers", "uitest: run ui-tests that get evaluated by humans"
    )
    config.addinivalue_line(
        "markers", "clear_testbed: create a testbed without any existing files"
    )
    config.addinivalue_line("markers", "no_git_testbed: create a testbed without git")


def pytest_collection_modifyitems(config, items):
    benchmark = config.getoption("--benchmark")
    uitest = config.getoption("--uitest")
    items[:] = filter_mark(items, "benchmark", benchmark)
    items[:] = filter_mark(items, "uitest", uitest)


@pytest.fixture
def get_marks(request):
    return [mark.name for mark in request.node.iter_markers()]


@pytest.fixture
def mock_collect_user_input(mocker):
    async_mock = AsyncMock()

    mocker.patch("mentat.session_input._get_input_request", side_effect=async_mock)

    def set_stream_messages(values):
        async_mock.side_effect = [
            StreamMessage(
                id=uuid4(),
                channel="default",
                source=StreamMessageSource.CLIENT,
                data=value,
                extra={},
                created_at=datetime.utcnow(),
            )
            for value in values
        ]

    async_mock.set_stream_messages = set_stream_messages
    return async_mock


@pytest.fixture(scope="function")
def mock_call_llm_api(mocker):
    completion_mock = mocker.patch.object(LlmApiHandler, "call_llm_api")

    def set_streamed_values(values):
        async def _async_generator():
            timestamp = int(time.time())
            for value in values:
                yield ChatCompletionChunk(
                    id="test-id",
                    choices=[
                        AsyncChoice(
                            delta=ChoiceDelta(content=value, role="assistant"),
                            finish_reason=None,
                            index=0,
                        )
                    ],
                    created=timestamp,
                    model="test-model",
                    object="chat.completion.chunk",
                )

        completion_mock.return_value = _async_generator()

    completion_mock.set_streamed_values = set_streamed_values

    def set_unstreamed_values(value):
        timestamp = int(time.time())
        completion_mock.return_value = ChatCompletion(
            id="test-id",
            choices=[
                Choice(
                    finish_reason="stop",
                    index=0,
                    message=ChatCompletionMessage(
                        content=value,
                        role="assistant",
                    ),
                )
            ],
            created=timestamp,
            model="test-model",
            object="chat.completion",
        )

    completion_mock.set_unstreamed_values = set_unstreamed_values
    return completion_mock


@pytest.fixture(scope="function")
def mock_call_embedding_api(mocker):
    embedding_mock = mocker.patch.object(LlmApiHandler, "call_embedding_api")

    def set_embedding_values(value):
        embedding_mock.return_value = value

    embedding_mock.set_embedding_values = set_embedding_values
    return embedding_mock


### Auto-used fixtures


@pytest.fixture(autouse=True, scope="function")
def mock_initialize_client(mocker, request):
    if not request.config.getoption("--benchmark"):
        mocker.patch.object(LlmApiHandler, "initialize_client")


# ContextVars need to be set in a synchronous fixture due to pytest not propagating
# async fixture contexts to test contexts.
# https://github.com/pytest-dev/pytest-asyncio/issues/127


@pytest.fixture(autouse=True)
def mock_session_context(temp_testbed):
    """
    This is autoused to make it easier to write tests without having to worry about whether
    or not SessionContext is set; however, this SessionContext will be overwritten by the SessionContext
    set by a Session if the test creates a Session.
    If you create a Session or Client in your test, do NOT use this SessionContext!
    """
    git_root = get_git_root_for_path(temp_testbed, raise_error=False)

    stream = SessionStream()

    cost_tracker = CostTracker()

    config = Config()

    llm_api_handler = LlmApiHandler()

    code_context = CodeContext(stream, git_root)

    code_file_manager = CodeFileManager()
    conversation = Conversation()

    vision_manager = VisionManager()

    agent_handler = AgentHandler()

    auto_completer = AutoCompleter()

    sampler = Sampler()

    session_context = SessionContext(
        Path.cwd(),
        stream,
        llm_api_handler,
        cost_tracker,
        config,
        code_context,
        code_file_manager,
        conversation,
        vision_manager,
        agent_handler,
        auto_completer,
        sampler,
    )
    token = SESSION_CONTEXT.set(session_context)
    yield session_context
    SESSION_CONTEXT.reset(token)


@pytest.fixture
def mock_code_context(temp_testbed, mock_session_context):
    return mock_session_context.code_context


### Auto-used fixtures


def run_git_command(cwd, *args):
    """Helper function to run a git command."""
    subprocess.run(
        ["git"] + list(args),
        cwd=cwd,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )


def add_permissions(func, path, exc_info):
    """
    Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.

    If the error is for another reason it re-raises the error.
    """

    gc.collect()  # Force garbage collection
    # Is the error an access error?
    if not os.access(path, os.W_OK):
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise


@pytest.fixture(autouse=True)
def temp_testbed(monkeypatch, get_marks):
    # Allow us to run tests from any directory
    base_dir = Path(__file__).parent.parent

    # create temporary copy of testbed, complete with git repo
    # realpath() resolves symlinks, required for paths to match on macOS
    temp_dir = os.path.realpath(tempfile.mkdtemp())
    temp_testbed = os.path.join(temp_dir, "testbed")
    os.mkdir(temp_testbed)

    if "no_git_testbed" not in get_marks:
        # Initialize git repo
        run_git_command(temp_testbed, "init")

        # Set local config for user.name and user.email. Set automatically on
        # MacOS, but not Windows/Ubuntu, which prevents commits from taking.
        run_git_command(temp_testbed, "config", "user.email", "test@example.com")
        run_git_command(temp_testbed, "config", "user.name", "Test User")

    if "clear_testbed" not in get_marks:
        # Copy testbed
        shutil.copytree(base_dir / "testbed", temp_testbed, dirs_exist_ok=True)
        shutil.copy(base_dir / ".gitignore", temp_testbed)

        if "no_git_testbed" not in get_marks:
            # Add all files and commit
            run_git_command(temp_testbed, "add", ".")
            run_git_command(temp_testbed, "commit", "-m", "add testbed")

    # necessary to undo chdir before calling rmtree, or it fails on windows
    with monkeypatch.context() as m:
        m.chdir(temp_testbed)
        yield Path(temp_testbed)

    shutil.rmtree(temp_dir, onerror=add_permissions)


# Always set the user config to just be a config in the temp_testbed; that way,
# it will be unset unless a specific test wants to make a config in the testbed
@pytest.fixture(autouse=True)
def mock_user_config(mocker):
    config.user_config_path = Path(config_file_name)


@pytest.fixture(autouse=True)
def mock_sleep_time(mocker):
    mocker.patch.object(StreamingPrinter, "sleep_time", new=lambda self: 0)



from typing import Iterable, List

from prompt_toolkit.completion import CompleteEvent, Completer
from prompt_toolkit.completion import Completion as PromptToolkitCompletion
from prompt_toolkit.document import Document

from mentat.auto_completer import Completion
from mentat.session_stream import SessionStream, StreamMessageSource


class MentatCompleter(Completer):
    def __init__(self, stream: SessionStream):
        self.stream = stream

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[PromptToolkitCompletion]:
        raise NotImplementedError

    async def get_completions_async(
        self, document: Document, complete_event: CompleteEvent
    ):
        text = document.text_before_cursor
        message = self.stream.send(
            text, source=StreamMessageSource.CLIENT, channel="completion_request"
        )

        response = await self.stream.recv(channel=f"completion_request:{message.id}")
        completions: List[Completion] = response.data
        for completion in completions:
            yield PromptToolkitCompletion(
                text=completion["content"],
                start_position=completion["position"],
                display=completion["display"],
            )



from tqdm import tqdm

from mentat.session_stream import StreamMessage


class LoadingHandler:
    def __init__(self):
        self.pbar = None

    def update(self, message: StreamMessage):
        """Create or update a loading bar with text and progress value (0-100)"""

        terminate = bool(message.extra and message.extra.get("terminate"))
        if self.pbar is None and terminate:
            self.terminate()
            return

        text = "" if not isinstance(message.data, str) else message.data
        if self.pbar is None:
            self.pbar = tqdm(
                total=100,
                desc=text,
                bar_format="{percentage:3.0f}%|{bar:50}| {desc}",
            )
        elif text:
            self.pbar.set_description(text)

        if "progress" in message.extra:
            _progress = min(message.extra["progress"], self.pbar.total - self.pbar.n)
            self.pbar.update(_progress)

        if terminate or self.pbar.n == self.pbar.total:
            self.terminate()

    def terminate(self, message: str | None = None):
        if self.pbar is not None:
            if message is not None:
                self.pbar.set_description(message)
            if self.pbar.total > self.pbar.n:
                self.pbar.update(self.pbar.total - self.pbar.n)
            self.pbar.close()
            self.pbar = None



from __future__ import annotations

import os
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Set, Union

from mentat.code_feature import (
    CodeFeature,
    get_code_message_from_features,
    get_consolidated_feature_refs,
    split_file_into_intervals,
)
from mentat.diff_context import DiffContext
from mentat.errors import PathValidationError, ReturnToUser
from mentat.feature_filters.default_filter import DefaultFilter
from mentat.feature_filters.embedding_similarity_filter import EmbeddingSimilarityFilter
from mentat.git_handler import get_paths_with_git_diffs
from mentat.include_files import (
    PathType,
    build_path_tree,
    get_code_features_for_path,
    get_path_type,
    get_paths_for_directory,
    is_file_text_encoded,
    match_path_with_patterns,
    print_path_tree,
    validate_and_format_path,
)
from mentat.interval import parse_intervals, split_intervals_from_path
from mentat.llm_api_handler import count_tokens, get_max_tokens, is_context_sufficient
from mentat.session_context import SESSION_CONTEXT
from mentat.session_stream import SessionStream


class CodeContext:
    def __init__(
        self,
        stream: SessionStream,
        git_root: Optional[Path] = None,
        diff: Optional[str] = None,
        pr_diff: Optional[str] = None,
        ignore_patterns: Iterable[Path | str] = [],
    ):
        self.git_root = git_root
        self.diff = diff
        self.pr_diff = pr_diff
        self.ignore_patterns = set(Path(p) for p in ignore_patterns)

        self.diff_context = None
        if self.git_root:
            self.diff_context = DiffContext(
                stream, self.git_root, self.diff, self.pr_diff
            )

        self.include_files: Dict[Path, List[CodeFeature]] = {}
        self.ignore_files: Set[Path] = set()
        self.auto_features: List[CodeFeature] = []

    def display_context(self):
        """Display the baseline context: included files and auto-context settings"""
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        config = session_context.config

        stream.send("Code Context:", style="code")
        prefix = "  "
        stream.send(f"{prefix}Directory: {session_context.cwd}")
        if self.diff_context and self.diff_context.name:
            stream.send(f"{prefix}Diff:", end=" ")
            stream.send(self.diff_context.get_display_context(), style="success")

        if config.auto_context_tokens > 0:
            stream.send(f"{prefix}Auto-Context: Enabled")
            stream.send(f"{prefix}Auto-Context Tokens: {config.auto_context_tokens}")
        else:
            stream.send(f"{prefix}Auto-Context: Disabled")

        if self.include_files:
            stream.send(f"{prefix}Included files:")
            stream.send(f"{prefix + prefix}{session_context.cwd.name}")
            features = [
                feature
                for file_features in self.include_files.values()
                for feature in file_features
            ]
            refs = get_consolidated_feature_refs(features)
            print_path_tree(
                build_path_tree([Path(r) for r in refs], session_context.cwd),
                get_paths_with_git_diffs(self.git_root) if self.git_root else set(),
                session_context.cwd,
                prefix + prefix,
            )
        else:
            stream.send(f"{prefix}Included files: ", end="")
            stream.send("None", style="warning")

        if self.auto_features:
            stream.send(f"{prefix}Auto-Included Features:")
            refs = get_consolidated_feature_refs(self.auto_features)
            print_path_tree(
                build_path_tree([Path(r) for r in refs], session_context.cwd),
                get_paths_with_git_diffs(self.git_root) if self.git_root else set(),
                session_context.cwd,
                prefix + prefix,
            )

    async def get_code_message(
        self,
        prompt_tokens: int,
        prompt: Optional[str] = None,
        expected_edits: Optional[list[str]] = None,  # for training/benchmarking
        loading_multiplier: float = 0.0,
    ) -> str:
        """
        Retrieves the current code message.
        'prompt' argument is embedded and used to search for similar files when auto-context is enabled.
        If prompt is empty, auto context won't be used.
        'prompt_tokens' argument is the total number of tokens used by the prompt before the code message,
        used to ensure that the code message won't overflow the model's context size
        """
        session_context = SESSION_CONTEXT.get()
        config = session_context.config
        model = config.model

        # Setup code message metadata
        code_message = list[str]()
        if self.diff_context:
            self.diff_context.clear_cache()
            if self.diff_context.diff_files():
                code_message += [
                    "Diff References:",
                    f' "-" = {self.diff_context.name}',
                    ' "+" = Active Changes',
                    "",
                ]

        code_message += ["Code Files:\n"]
        meta_tokens = count_tokens("\n".join(code_message), model, full_message=True)

        # Calculate user included features token size
        include_features = [
            feature
            for file_features in self.include_files.values()
            for feature in file_features
        ]
        include_files_message = get_code_message_from_features(include_features)
        include_files_tokens = count_tokens(
            "\n".join(include_files_message), model, full_message=False
        )

        tokens_used = prompt_tokens + meta_tokens + include_files_tokens
        if not is_context_sufficient(tokens_used):
            raise ReturnToUser()
        auto_tokens = min(
            get_max_tokens() - tokens_used - config.token_buffer,
            config.auto_context_tokens,
        )

        # Get auto included features
        if config.auto_context_tokens > 0 and prompt:
            features = self.get_all_features()
            feature_filter = DefaultFilter(
                auto_tokens,
                prompt,
                expected_edits,
                loading_multiplier=loading_multiplier,
            )
            self.auto_features = list(
                set(self.auto_features) | set(await feature_filter.filter(features))
            )

        # Merge include file features and auto features and add to code message
        code_message += get_code_message_from_features(
            include_features + self.auto_features
        )

        return "\n".join(code_message)

    def get_all_features(
        self,
        max_chars: int = 100000,
        split_intervals: bool = True,
    ) -> list[CodeFeature]:
        """
        Retrieves every CodeFeature under the cwd. If files_only is True the features won't be split into intervals
        """
        session_context = SESSION_CONTEXT.get()

        abs_exclude_patterns: Set[Path] = set()
        for pattern in self.ignore_patterns.union(
            session_context.config.file_exclude_glob_list
        ):
            if not Path(pattern).is_absolute():
                abs_exclude_patterns.add(session_context.cwd / pattern)
            else:
                abs_exclude_patterns.add(Path(pattern))

        all_features: List[CodeFeature] = []
        for path in get_paths_for_directory(
            path=session_context.cwd, exclude_patterns=abs_exclude_patterns
        ):
            if not is_file_text_encoded(path) or os.path.getsize(path) > max_chars:
                continue

            if not split_intervals:
                _feature = CodeFeature(path)
                all_features.append(_feature)
            else:
                full_feature = CodeFeature(path)
                _split_features = split_file_into_intervals(full_feature)
                all_features += _split_features

        return sorted(all_features, key=lambda f: f.path)

    def clear_auto_context(self):
        """
        Clears all auto-features added to the conversation so far.
        """
        self.auto_features = []

    def include_features(self, code_features: Iterable[CodeFeature]):
        """
        Adds the given code features to context. If the feature is already included, it will not be added.
        """
        included_paths: Set[Path] = set()
        for code_feature in code_features:
            if code_feature.path not in self.include_files:
                self.include_files[code_feature.path] = [code_feature]
                included_paths.add(Path(str(code_feature)))
            else:
                code_feature_not_included = True
                for included_code_feature in self.include_files[code_feature.path]:
                    # Intervals can still overlap if user includes intervals different than what ctags breaks up,
                    # but we merge when making code message and don't duplicate lines
                    if (
                        included_code_feature.interval == code_feature.interval
                        # No need to include an interval if the entire file is already included
                        or included_code_feature.interval.whole_file()
                    ):
                        code_feature_not_included = False
                        break
                if code_feature_not_included:
                    if code_feature.interval.whole_file():
                        self.include_files[code_feature.path] = []
                    self.include_files[code_feature.path].append(code_feature)
                    included_paths.add(Path(str(code_feature)))
        return included_paths

    def include(
        self, path: Path | str, exclude_patterns: Iterable[Path | str] = []
    ) -> Set[Path]:
        """
        Add paths to the context

        Paths that are already in the context and invalid paths are ignored.

        The following are behaviors for each `path` type:

        - File: the file is added to the context.
        - File interval: the file is added to the context, and the interval is added to the file
        - Directory: all files in the directory are recursively added to the context
        - Glob pattern: all files that match the glob pattern are added to the context

        Args:
            `path`: can be a relative or absolute file path, file interval path, directory, or glob pattern.

        Return:
            A set of paths that have been successfully included in the context
        """
        session_context = SESSION_CONTEXT.get()

        path = Path(path)

        abs_exclude_patterns: Set[Path] = set()
        all_exclude_patterns: Set[Union[str, Path]] = set(
            [
                *exclude_patterns,
                *self.ignore_patterns,
                *session_context.config.file_exclude_glob_list,
            ]
        )
        for pattern in all_exclude_patterns:
            if not Path(pattern).is_absolute():
                abs_exclude_patterns.add(session_context.cwd / pattern)
            else:
                abs_exclude_patterns.add(Path(pattern))

        try:
            code_features = get_code_features_for_path(
                path=path,
                cwd=session_context.cwd,
                exclude_patterns=abs_exclude_patterns,
            )
        except PathValidationError as e:
            session_context.stream.send(str(e), style="error")
            return set()

        return self.include_features(code_features)

    def _exclude_file(self, path: Path) -> Path | None:
        session_context = SESSION_CONTEXT.get()
        if path in self.include_files:
            del self.include_files[path]
            return path
        else:
            session_context.stream.send(f"Path {path} not in context", style="error")

    def _exclude_file_interval(self, path: Path) -> Set[Path]:
        session_context = SESSION_CONTEXT.get()

        excluded_paths: Set[Path] = set()

        interval_path, interval_str = split_intervals_from_path(path)
        if interval_path not in self.include_files:
            session_context.stream.send(
                f"Path {interval_path} not in context", style="error"
            )
            return excluded_paths

        intervals = parse_intervals(interval_str)
        included_code_features: List[CodeFeature] = []
        for code_feature in self.include_files[interval_path]:
            if code_feature.interval not in intervals:
                included_code_features.append(code_feature)
            else:
                excluded_paths.add(Path(str(code_feature)))

        if len(included_code_features) == 0:
            del self.include_files[interval_path]
        else:
            self.include_files[interval_path] = included_code_features

        return excluded_paths

    def _exclude_directory(self, path: Path) -> Set[Path]:
        excluded_paths: Set[Path] = set()

        paths_to_exclude: Set[Path] = set()
        for included_path in self.include_files:
            if path in included_path.parents:
                paths_to_exclude.add(included_path)
        for excluded_path in paths_to_exclude:
            del self.include_files[excluded_path]
            excluded_paths.add(excluded_path)

        return excluded_paths

    def _exclude_glob(self, path: Path) -> Set[Path]:
        excluded_paths: Set[Path] = set()

        paths_to_exclude: Set[Path] = set()
        for included_path in self.include_files:
            if match_path_with_patterns(included_path, set([path])):
                paths_to_exclude.add(included_path)
        for excluded_path in paths_to_exclude:
            del self.include_files[excluded_path]
            excluded_paths.add(excluded_path)

        return excluded_paths

    def exclude(self, path: Path | str) -> Set[Path]:
        """Remove code from the context

        Paths that are not in the context and invalid paths are ignored.

        The following are behaviors for each `path` type:

        - File: the file is removed from the context
        - File interval: the interval is removed from the file. If the file doesn't have the exact interval specified,
          nothing is removed.
        - Directory: all files in the directory are removed from the context
        - Glob pattern: all files that match the glob pattern are removed from the context

        Args:
            `path`: can be a relative or absolute file path, file interval path, directory, or glob pattern.

        Return:
            A set of paths that have been successfully excluded from the context
        """
        session_context = SESSION_CONTEXT.get()

        path = Path(path)
        excluded_paths: Set[Path] = set()
        try:
            validated_path = validate_and_format_path(
                path, session_context.cwd, check_for_text=False
            )
            match get_path_type(validated_path):
                case PathType.FILE:
                    excluded_path = self._exclude_file(validated_path)
                    if excluded_path:
                        excluded_paths.add(excluded_path)
                case PathType.FILE_INTERVAL:
                    excluded_paths.update(self._exclude_file_interval(validated_path))
                case PathType.DIRECTORY:
                    excluded_paths.update(self._exclude_directory(validated_path))
                case PathType.GLOB:
                    excluded_paths.update(self._exclude_glob(validated_path))
        except PathValidationError as e:
            session_context.stream.send(str(e), style="error")

        return excluded_paths

    async def search(
        self,
        query: str,
        max_results: int | None = None,
    ) -> list[tuple[CodeFeature, float]]:
        """Return the top n features that are most similar to the query."""

        all_features = self.get_all_features()

        embedding_similarity_filter = EmbeddingSimilarityFilter(query)
        all_features_sorted = await embedding_similarity_filter.score(all_features)
        if max_results is None:
            return all_features_sorted
        else:
            return all_features_sorted[:max_results]



import asyncio
import logging
import os
import traceback
from asyncio import CancelledError, Task
from pathlib import Path
from typing import Any, Coroutine, List, Optional, Set
from uuid import uuid4

import attr
import sentry_sdk
from openai import APITimeoutError, BadRequestError, RateLimitError

from mentat.agent_handler import AgentHandler
from mentat.auto_completer import AutoCompleter
from mentat.code_context import CodeContext
from mentat.code_edit_feedback import get_user_feedback_on_edits
from mentat.code_file_manager import CodeFileManager
from mentat.config import Config
from mentat.conversation import Conversation
from mentat.cost_tracker import CostTracker
from mentat.ctags import ensure_ctags_installed
from mentat.errors import MentatError, ReturnToUser, SampleError, SessionExit, UserError
from mentat.git_handler import get_git_root_for_path
from mentat.llm_api_handler import LlmApiHandler, is_test_environment
from mentat.logging_config import setup_logging
from mentat.sampler.sampler import Sampler
from mentat.sentry import sentry_init
from mentat.session_context import SESSION_CONTEXT, SessionContext
from mentat.session_input import collect_input_with_commands
from mentat.session_stream import SessionStream
from mentat.utils import check_version, mentat_dir_path
from mentat.vision.vision_manager import VisionManager


class Session:
    """
    The server for Mentat.
    To stop, send a message on the session_exit channel.
    A message will be sent on the client_exit channel when ready for client to quit.
    """

    def __init__(
        self,
        cwd: Path,
        paths: List[Path] = [],
        exclude_paths: List[Path] = [],
        ignore_paths: List[Path] = [],
        diff: Optional[str] = None,
        pr_diff: Optional[str] = None,
        config: Config = Config(),
    ):
        # All errors thrown here need to be caught here
        self.stopped = False

        if not mentat_dir_path.exists():
            os.mkdir(mentat_dir_path)
        setup_logging()
        sentry_init()
        self.id = uuid4()
        self._tasks: Set[asyncio.Task[None]] = set()

        # Since we can't set the session_context until after all of the singletons are created,
        # any singletons used in the constructor of another singleton must be passed in
        git_root = get_git_root_for_path(cwd, raise_error=False)

        llm_api_handler = LlmApiHandler()

        stream = SessionStream()
        self.stream = stream
        self.stream.start()

        cost_tracker = CostTracker()

        code_context = CodeContext(stream, git_root, diff, pr_diff, ignore_paths)

        code_file_manager = CodeFileManager()

        conversation = Conversation()

        vision_manager = VisionManager()

        agent_handler = AgentHandler()

        auto_completer = AutoCompleter()

        sampler = Sampler()

        session_context = SessionContext(
            cwd,
            stream,
            llm_api_handler,
            cost_tracker,
            config,
            code_context,
            code_file_manager,
            conversation,
            vision_manager,
            agent_handler,
            auto_completer,
            sampler,
        )
        self.ctx = session_context
        SESSION_CONTEXT.set(session_context)
        self.error = None

        # Functions that require session_context
        check_version()
        config.send_errors_to_stream()
        for path in paths:
            code_context.include(path, exclude_patterns=exclude_paths)
        if (
            code_context.diff_context is not None
            and len(code_context.include_files) == 0
            and (diff or pr_diff)
        ):
            for file in code_context.diff_context.diff_files():
                code_context.include(file)

    def _create_task(self, coro: Coroutine[None, None, Any]):
        """Utility method for running a Task in the background"""

        def task_cleanup(task: asyncio.Task[None]):
            self._tasks.remove(task)

        task = asyncio.create_task(coro)
        task.add_done_callback(task_cleanup)
        self._tasks.add(task)

        return task

    async def _main(self):
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        code_context = session_context.code_context
        conversation = session_context.conversation
        code_file_manager = session_context.code_file_manager
        agent_handler = session_context.agent_handler

        # check early for ctags so we can fail fast
        if session_context.config.auto_context_tokens > 0:
            ensure_ctags_installed()

        session_context.llm_api_handler.initialize_client()
        code_context.display_context()
        await conversation.display_token_count()

        stream.send("Type 'q' or use Ctrl-C to quit at any time.")
        need_user_request = True
        while True:
            try:
                if need_user_request:
                    # Normally, the code_file_manager pushes the edits; but when agent mode is on, we want all
                    # edits made between user input to be collected together.
                    if agent_handler.agent_enabled:
                        code_file_manager.history.push_edits()
                        stream.send(
                            "Use /undo to undo all changes from agent mode since last"
                            " input.",
                            style="success",
                        )
                    stream.send("\nWhat can I do for you?", style="input")
                    message = await collect_input_with_commands()
                    if message.data.strip() == "":
                        continue
                    conversation.add_user_message(message.data)

                parsed_llm_response = await conversation.get_model_response()
                file_edits = [
                    file_edit
                    for file_edit in parsed_llm_response.file_edits
                    if file_edit.is_valid()
                ]
                if file_edits:
                    if not agent_handler.agent_enabled:
                        file_edits, need_user_request = (
                            await get_user_feedback_on_edits(file_edits)
                        )
                    for file_edit in file_edits:
                        file_edit.resolve_conflicts()

                    if session_context.sampler and session_context.sampler.active:
                        try:
                            session_context.sampler.set_active_diff()
                        except SampleError as e:
                            stream.send(
                                f"Sampler error setting active diff: {e}. Disabling"
                                " sampler.",
                                style="error",
                            )
                            session_context.sampler.active = False

                    applied_edits = await code_file_manager.write_changes_to_files(
                        file_edits
                    )
                    stream.send(
                        "Changes applied." if applied_edits else "No changes applied.",
                        style="input",
                    )

                    if agent_handler.agent_enabled:
                        if parsed_llm_response.interrupted:
                            need_user_request = True
                        else:
                            need_user_request = await agent_handler.add_agent_context()
                else:
                    need_user_request = True
                stream.send(bool(file_edits), channel="edits_complete")
            except SessionExit:
                break
            except ReturnToUser:
                need_user_request = True
                continue
            except (APITimeoutError, RateLimitError, BadRequestError) as e:
                stream.send(f"Error accessing OpenAI API: {e.message}", style="error")
                break

    async def listen_for_session_exit(self):
        await self.stream.recv(channel="session_exit")
        self._main_task.cancel()

    async def listen_for_completion_requests(self):
        ctx = SESSION_CONTEXT.get()

        async for message in self.stream.listen(channel="completion_request"):
            completions = ctx.auto_completer.get_completions(message.data)
            # Will intermediary client for vscode serialize/deserialize all messages automatically?
            self.stream.send(completions, channel=f"completion_request:{message.id}")

    ### lifecycle

    def start(self):
        """Asynchronously start the Session.

        A background asyncio Task will be created to run the startup sequence and run
        the main loop which runs until an Exception or session_exit signal is encountered.
        """

        async def run_main():
            ctx = SESSION_CONTEXT.get()
            try:
                with sentry_sdk.start_transaction(
                    op="mentat_started", name="Mentat Started"
                ) as transaction:
                    transaction.set_tag("config", attr.asdict(ctx.config))
                    await self._main()
            except (SessionExit, CancelledError):
                pass
            except (MentatError, UserError) as e:
                self.stream.send(str(e), style="error")
            except Exception as e:
                # All unhandled exceptions end up here
                error = f"Unhandled Exception: {traceback.format_exc()}"
                # Helps us handle errors in tests
                if is_test_environment():
                    print(error)
                self.error = error
                sentry_sdk.capture_exception(e)
                self.stream.send(error, style="error")
            finally:
                await self._stop()
                sentry_sdk.flush()

        self._main_task: Task[None] = asyncio.create_task(run_main())

        self._create_task(self.listen_for_session_exit())
        self._create_task(self.listen_for_completion_requests())

    async def _stop(self):
        if self.stopped:
            return
        self.stopped = True

        session_context = SESSION_CONTEXT.get()
        cost_tracker = session_context.cost_tracker
        vision_manager = session_context.vision_manager

        vision_manager.close()
        cost_tracker.display_total_cost()
        logging.shutdown()

        for task in self._tasks:
            task.cancel()

        self._main_task.cancel()
        try:
            await self._main_task
        except CancelledError:
            pass

        self.stream.send(None, channel="client_exit")
        await self.stream.join()
        self.stream.stop()



from __future__ import annotations

import asyncio
import json
import logging
import subprocess
from typing import List, Optional

from openai import RateLimitError
from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionContentPartParam,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionUserMessageParam,
)

from mentat.llm_api_handler import (
    TOKEN_COUNT_WARNING,
    count_tokens,
    get_max_tokens,
    prompt_tokens,
)
from mentat.parsers.file_edit import FileEdit
from mentat.parsers.parser import ParsedLLMResponse
from mentat.session_context import SESSION_CONTEXT
from mentat.transcripts import ModelMessage, TranscriptMessage, UserMessage
from mentat.utils import add_newline


class Conversation:
    def __init__(self):
        self._messages = list[ChatCompletionMessageParam]()

        # This contains a list of messages used for transcripts
        self.literal_messages = list[TranscriptMessage]()

    async def display_token_count(self):
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        config = session_context.config
        code_context = session_context.code_context

        if "gpt-4" not in config.model:
            stream.send(
                "Warning: Mentat has only been tested on GPT-4. You may experience"
                " issues with quality. This model may not be able to respond in"
                " mentat's edit format.",
                style="warning",
            )
            if "gpt-3.5" not in config.model:
                stream.send(
                    "Warning: Mentat does not know how to calculate costs or context"
                    " size for this model.",
                    style="warning",
                )

        messages = self.get_messages()
        code_message = await code_context.get_code_message(
            prompt_tokens(
                messages,
                config.model,
            )
        )
        messages.append(
            ChatCompletionSystemMessageParam(
                role="system",
                content=code_message,
            )
        )
        tokens = prompt_tokens(messages, config.model)

        context_size = get_max_tokens()
        if tokens + config.token_buffer > context_size:
            _plural = len(code_context.include_files) > 1
            _exceed = tokens > context_size
            message: dict[tuple[bool, bool], str] = {
                (False, False): " is close to",
                (False, True): " exceeds",
                (True, False): "s are close to",
                (True, True): "s exceed",
            }
            stream.send(
                f"Included file{message[(_plural, _exceed)]} token limit"
                f" ({tokens} / {context_size}). Truncating based on task similarity.",
                style="warning",
            )
        else:
            stream.send(
                f"Prompt and included files token count: {tokens} / {context_size}",
                style="info",
            )

    # The transcript logger logs tuples containing the actual message sent by the user or LLM
    # and (for LLM messages) the LLM conversation that led to that LLM response
    def add_transcript_message(self, transcript_message: TranscriptMessage):
        transcript_logger = logging.getLogger("transcript")
        transcript_logger.info(json.dumps(transcript_message))
        self.literal_messages.append(transcript_message)

    def add_user_message(self, message: str, image: Optional[str] = None):
        """Used for actual user input messages"""
        content: List[ChatCompletionContentPartParam] | str = message
        if image:
            content = [
                {
                    "type": "text",
                    "text": message,
                },
                {
                    "type": "image_url",
                    "image_url": {
                        "url": image,
                    },
                },
            ]
        self.add_transcript_message(UserMessage(message=content, prior_messages=None))
        self.add_message(ChatCompletionUserMessageParam(role="user", content=content))

    def add_model_message(
        self, message: str, messages_snapshot: list[ChatCompletionMessageParam]
    ):
        """Used for actual model output messages"""
        self.add_transcript_message(
            ModelMessage(message=message, prior_messages=messages_snapshot)
        )
        self.add_message(
            ChatCompletionAssistantMessageParam(role="assistant", content=message)
        )

    def add_message(self, message: ChatCompletionMessageParam):
        """Used for adding messages to the models conversation. Does not add a left-side message to the transcript!"""
        self._messages.append(message)

    def get_messages(
        self, include_system_prompt: bool = True
    ) -> list[ChatCompletionMessageParam]:
        """Returns the messages in the conversation. The system message may change throughout
        the conversation so it is important to access the messages through this method.
        """
        session_context = SESSION_CONTEXT.get()
        config = session_context.config
        if config.no_parser_prompt or not include_system_prompt:
            return self._messages.copy()
        else:
            parser = config.parser
            prompt = parser.get_system_prompt()
            prompt_message: ChatCompletionMessageParam = (
                ChatCompletionSystemMessageParam(
                    role="system",
                    content=prompt,
                )
            )
            return [prompt_message] + self._messages.copy()

    def clear_messages(self) -> None:
        """Clears the messages in the conversation"""
        self._messages = list[ChatCompletionMessageParam]()

    async def _stream_model_response(
        self,
        messages: list[ChatCompletionMessageParam],
        loading_multiplier: float = 0.0,
    ) -> ParsedLLMResponse:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        config = session_context.config
        parser = config.parser
        llm_api_handler = session_context.llm_api_handler
        cost_tracker = session_context.cost_tracker

        if loading_multiplier:
            stream.send(
                "Sending query and context to LLM",
                channel="loading",
                progress=50 * loading_multiplier,
            )
        response = await llm_api_handler.call_llm_api(
            messages,
            config.model,
            stream=True,
            response_format=parser.response_format(),
        )
        if loading_multiplier:
            stream.send(
                None,
                channel="loading",
                progress=50 * loading_multiplier,
                terminate=True,
            )

        num_prompt_tokens = prompt_tokens(messages, config.model)
        stream.send(f"Total token count: {num_prompt_tokens}", style="info")
        if num_prompt_tokens > TOKEN_COUNT_WARNING:
            stream.send(
                "Warning: LLM performance drops off rapidly at large context sizes. Use"
                " /clear to clear context or use /exclude to exclude any uneccessary"
                " files.",
                style="warning",
            )

        stream.send("Streaming... use control-c to interrupt the model at any point\n")
        async with parser.interrupt_catcher():
            parsed_llm_response = await parser.stream_and_parse_llm_response(
                add_newline(response)
            )
        if not parsed_llm_response.interrupted:
            cost_tracker.display_last_api_call()
        else:
            # Generator doesn't log the api call if we interrupt it
            cost_tracker.log_api_call_stats(
                num_prompt_tokens,
                count_tokens(
                    parsed_llm_response.full_response, config.model, full_message=False
                ),
                config.model,
                display=True,
            )

        messages.append(
            ChatCompletionAssistantMessageParam(
                role="assistant", content=parsed_llm_response.full_response
            )
        )
        self.add_model_message(parsed_llm_response.full_response, messages)

        return parsed_llm_response

    async def get_model_response(self) -> ParsedLLMResponse:
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        config = session_context.config
        code_context = session_context.code_context

        messages_snapshot = self.get_messages()

        # Get current code message
        loading_multiplier = 1.0 if config.auto_context_tokens > 0 else 0.0
        prompt = messages_snapshot[-1]["content"]
        if isinstance(prompt, list):
            text_prompts = [
                p.get("text", "") for p in prompt if p.get("type") == "text"
            ]
            prompt = " ".join(text_prompts)
        code_message = await code_context.get_code_message(
            prompt_tokens(messages_snapshot, config.model),
            prompt=(
                prompt  # Prompt can be image as well as text
                if isinstance(prompt, str)
                else ""
            ),
            loading_multiplier=0.5 * loading_multiplier,
        )
        messages_snapshot.insert(
            1, ChatCompletionSystemMessageParam(role="system", content=code_message)
        )

        try:
            response = await self._stream_model_response(
                messages_snapshot,
                loading_multiplier=0.5 * loading_multiplier,
            )
        except RateLimitError:
            stream.send(
                "Rate limit error received from OpenAI's servers using model"
                f' {config.model}.\nUse "/config model <model_name>" to switch to a'
                " different model.",
                style="error",
            )
            return ParsedLLMResponse("", "", list[FileEdit]())
        finally:
            if loading_multiplier:
                stream.send(None, channel="loading", terminate=True)
        return response

    def remaining_context(self) -> int | None:
        ctx = SESSION_CONTEXT.get()
        return get_max_tokens() - prompt_tokens(self.get_messages(), ctx.config.model)

    def can_add_to_context(self, message: str) -> bool:
        """
        Whether or not the model has enough context remaining to add this message.
        Will take token buffer into account and uses full_message=True.
        """
        ctx = SESSION_CONTEXT.get()

        remaining_context = self.remaining_context()
        return (
            remaining_context is not None
            and remaining_context
            - count_tokens(message, ctx.config.model, full_message=True)
            - ctx.config.token_buffer
            > 0
        )

    async def run_command(self, command: list[str]) -> bool:
        """
        Runs a command and, if there is room, adds the output to the conversation under the 'system' role.
        """
        ctx = SESSION_CONTEXT.get()

        ctx.stream.send("Running command: ", end="", style="info")
        ctx.stream.send(" ".join(command), style="warning")
        ctx.stream.send("Command output:", style="info")

        try:
            process = subprocess.Popen(
                command,
                cwd=ctx.cwd,
                text=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            output = list[str]()
            while True:
                if process.stdout is None:
                    break
                line = process.stdout.readline()
                if not line:
                    break
                output.append(line)
                ctx.stream.send(line, end="")
                # This gives control back to the asyncio event loop so we can actually print what we sent
                # Unfortunately asyncio.sleep(0) won't work https://stackoverflow.com/a/74505785
                # Note: if subprocess doesn't flush, output can't and won't be streamed.
                await asyncio.sleep(0.01)
        except FileNotFoundError:
            output = [f"Invalid command: {' '.join(command)}"]
            ctx.stream.send(output[0])
        output = "".join(output)
        message = f"Command ran:\n{' '.join(command)}\nCommand output:\n{output}"

        if self.can_add_to_context(message):
            self.add_message(
                ChatCompletionSystemMessageParam(role="system", content=message)
            )
            ctx.stream.send(
                "Successfully added command output to model context.", style="success"
            )
            return True
        else:
            ctx.stream.send(
                "Not enough tokens remaining in model's context to add command output"
                " to model context.",
                style="error",
            )
            return False



import subprocess
from pathlib import Path
from uuid import uuid4

from git import GitCommandError, Repo  # type: ignore
from openai.types.chat import ChatCompletionMessageParam

from mentat.code_feature import get_consolidated_feature_refs
from mentat.errors import SampleError
from mentat.git_handler import get_git_diff, get_git_root_for_path, get_hexsha_active
from mentat.parsers.git_parser import GitParser
from mentat.sampler.sample import Sample
from mentat.sampler.utils import get_active_snapshot_commit
from mentat.session_context import SESSION_CONTEXT
from mentat.session_input import collect_user_input
from mentat.utils import get_relative_path


def parse_message(message: ChatCompletionMessageParam) -> dict[str, str]:
    ctx = SESSION_CONTEXT.get()
    content = message.get("content")
    text, code = "", ""
    if isinstance(content, str):
        if message.get("role") != "assistant":
            text = content
        output = list[str]()
        in_special = False
        for line in content.splitlines():
            if ctx.config.parser._starts_special(line):  # type: ignore
                in_special = True
            if not in_special:
                output.append(line)
            else:
                pass  # TODO: Convert to git diff format, replace 'code' above
            if ctx.config.parser._ends_code(line):  # type: ignore
                in_special = False
        while output[-1] == "":
            output.pop()
        text = "\n".join(output)
    elif isinstance(content, list) and len(content) > 0:
        content = content[0]
        if "text" in content and isinstance(content.get("text"), str):  # type: ignore
            text = content.get("text")  # type: ignore
    return {"text": text, "code": code}


class Sampler:
    active: bool = True
    diff_active: str | None = None
    commit_active: str | None = None
    last_sample_id: str | None = None
    last_sample_hexsha: str | None = None

    def set_active_diff(self):
        # Create a temporary commit with the active changes
        ctx = SESSION_CONTEXT.get()
        git_root = get_git_root_for_path(ctx.cwd, raise_error=False)
        if not git_root:
            return
        repo = Repo(git_root)
        self.commit_active = get_active_snapshot_commit(repo)
        # If changes were made since the last sample, don't list it as parent.
        if not self.last_sample_hexsha:
            return
        if self.last_sample_hexsha != get_hexsha_active():
            self.last_sample_id = None
            self.last_sample_hexsha = None

    async def create_sample(self) -> Sample:
        # Check for repo and merge_base in config
        session_context = SESSION_CONTEXT.get()
        stream = session_context.stream
        code_context = session_context.code_context
        config = session_context.config
        conversation = session_context.conversation

        git_root = get_git_root_for_path(session_context.cwd, raise_error=False)
        if not git_root:
            raise SampleError("No git repo found")

        stream.send("Input sample data", style="input")
        git_repo = Repo(git_root)
        merge_base = None
        if config.sample_merge_base_target:
            target = config.sample_merge_base_target
            stream.send(f"Use merge base target from config ({target})? (y/N)")
            response = (await collect_user_input()).data.strip()
            if response == "y":
                try:
                    mb = git_repo.merge_base(git_repo.head.commit, target)[0]
                    assert mb and hasattr(mb, "hexsha"), "No merge base found"
                    merge_base = mb.hexsha
                except Exception as e:
                    stream.send(
                        f"Error getting merge base from tar: {e}", style="error"
                    )
        if not merge_base:
            merge_base = git_repo.head.commit.hexsha
            stream.send(
                f"Use latest commit ({merge_base[:10]} as merge base? Press 'ENTER' to"
                " accept, or enter a new merge base commit."
            )
            response = (await collect_user_input()).data.strip()
            if response:
                merge_base = response
        try:
            assert merge_base is not None, "No merge base found"
            diff_merge_base = get_git_diff(merge_base, "HEAD")
        except (AssertionError, GitCommandError) as e:
            raise SampleError(f"Error getting diff for merge base: {e}")

        repo = config.sample_repo
        if not repo:
            remote_url = ""
            try:
                remote_url = subprocess.check_output(
                    ["git", "config", "--get", "remote.origin.url"],
                    text=True,
                ).strip()
            except subprocess.CalledProcessError:
                pass
            stream.send(
                f"Found repo URL: {remote_url}. Press 'ENTER' to accept, or enter a new"
                " URL."
            )
            response = (await collect_user_input()).data.strip()
            if response == "y":
                repo = remote_url
            else:
                repo = response
            config.sample_repo = repo

        stream.send("Sample Title:")
        title = (await collect_user_input()).data.strip() or ""
        stream.send("Description: (optional)")
        description = (await collect_user_input()).data.strip() or ""
        stream.send("Test Command: (optional, e.g. 'pytest -k foo')")
        test_command = (await collect_user_input()).data.strip() or ""

        message_history, message_prompt, message_edit = list[dict[str, str]](), "", ""
        for m in conversation.get_messages()[::-1]:
            if m["role"] not in {"user", "assistant"}:
                continue
            parsed = parse_message(m)
            text = parsed["text"]
            code = parsed["code"]
            if not message_prompt:
                if m["role"] == "user":
                    message_prompt = text
                elif m["role"] == "assistant":
                    message_edit += text
            else:
                message_history.append({"role": m["role"], "content": text + code})

        diff_active = ""
        diff_edit = get_git_diff("HEAD")
        if self.commit_active:
            diff_active = get_git_diff("HEAD", self.commit_active)
            diff_edit = get_git_diff(self.commit_active)

        context = set[str]()

        def _rp(f: str | Path) -> str:
            return get_relative_path(Path(f), git_root).as_posix()

        # Add include_files from context
        if code_context.include_files:
            feature_refs = get_consolidated_feature_refs(
                [f for fs in code_context.include_files.values() for f in fs]
            )
            context.update(_rp(f) for f in feature_refs)
        # Undo adds/removes/renames to match pre-diff_edit state
        if diff_edit:
            file_edits = GitParser().parse_string(diff_edit).file_edits
            for file_edit in file_edits:
                file_path = _rp(file_edit.file_path)
                rename_path = (
                    ""
                    if not file_edit.rename_file_path
                    else _rp(file_edit.rename_file_path)
                )
                if file_edit.is_deletion or rename_path or file_edit.replacements:
                    context.add(file_path)
                if file_edit.is_creation and file_path in context:
                    context.remove(file_path)
                if rename_path and rename_path in context:
                    context.remove(rename_path)
            # TODO: Prompt User to modify/approve context

        sample = Sample(
            title=title,
            description=description,
            id=uuid4().hex,
            parent_id=self.last_sample_id or "",
            repo=repo,
            merge_base=merge_base,
            diff_merge_base=diff_merge_base,
            diff_active=diff_active,
            message_history=message_history,
            message_prompt=message_prompt,
            message_edit=message_edit,
            context=list(context),
            diff_edit=diff_edit,
            test_command=test_command,
        )

        # Save the hexsha and id
        self.last_sample_id = sample.id
        self.last_sample_hexsha = get_hexsha_active()
        self.commit_active = None

        return sample



from typing import Any

from prompt_toolkit import PromptSession
from prompt_toolkit.application.current import get_app
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory, Suggestion
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.filters import Condition
from prompt_toolkit.formatted_text import AnyFormattedText
from prompt_toolkit.history import FileHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding.key_processor import KeyPressEvent

from mentat.utils import mentat_dir_path


class FilteredFileHistory(FileHistory):
    def __init__(self, filename: str):
        self.excluded_phrases = ["y", "n", "i", "q"]
        super().__init__(filename)

    def append_string(self, string: str):
        if string.strip().lower() not in self.excluded_phrases and string.strip():
            super().append_string(string)


class FilteredHistorySuggestions(AutoSuggestFromHistory):
    def __init__(self):
        super().__init__()

    def get_suggestion(self, buffer: Buffer, document: Document) -> Suggestion | None:
        # We want the auto completer to handle commands instead of the suggester
        if buffer.text[0] == "/":
            return None
        else:
            return super().get_suggestion(buffer, document)


class MentatPromptSession(PromptSession[str]):
    def __init__(self, *args: Any, **kwargs: Any):
        self._setup_bindings()
        super().__init__(
            message=[("class:prompt", ">>> ")],
            history=FilteredFileHistory(str(mentat_dir_path.joinpath("history"))),
            auto_suggest=FilteredHistorySuggestions(),
            multiline=True,
            prompt_continuation=self.prompt_continuation,
            key_bindings=self.bindings,
            *args,
            **kwargs,
        )

    def prompt_continuation(
        self, width: int, line_number: int, is_soft_wrap: int
    ) -> AnyFormattedText:
        return (
            "" if is_soft_wrap else [("class:continuation", " " * (width - 2) + "> ")]
        )

    def _setup_bindings(self):
        self.bindings = KeyBindings()

        @Condition
        def not_searching() -> bool:
            return not get_app().layout.is_searching

        @self.bindings.add("s-down", filter=not_searching)
        @self.bindings.add("c-j", filter=not_searching)
        def _(event: KeyPressEvent):
            event.current_buffer.insert_text("\n")

        @self.bindings.add("enter", filter=not_searching)
        def _(event: KeyPressEvent):
            event.current_buffer.validate_and_handle()

        @Condition
        def complete_suggestion() -> bool:
            app = get_app()
            return (
                app.current_buffer.suggestion is not None
                and len(app.current_buffer.suggestion.text) > 0
                and app.current_buffer.document.is_cursor_at_the_end
                and bool(app.current_buffer.text)
                and app.current_buffer.text[0] != "/"
            )

        @self.bindings.add("right", filter=complete_suggestion)
        def _(event: KeyPressEvent):
            suggestion = event.current_buffer.suggestion
            if suggestion:
                event.current_buffer.insert_text(suggestion.text)

        @self.bindings.add("c-c", filter=not_searching)
        @self.bindings.add("c-d", filter=not_searching)
        def _(event: KeyPressEvent):
            if event.current_buffer.text != "":
                event.current_buffer.reset()
            else:
                event.app.exit(result="q")



from __future__ import annotations

import json
from pathlib import Path

import attr

from mentat.errors import SampleError
from mentat.sampler import __version__


@attr.define
class Sample:
    # TODO: enforce required fields
    title: str = attr.field(default="")
    description: str = attr.field(default="")
    id: str = attr.field(default="")
    parent_id: str = attr.field(default="")
    repo: str = attr.field(default="")
    merge_base: str | None = attr.field(default=None)
    diff_merge_base: str = attr.field(default="")
    diff_active: str = attr.field(default="")
    message_history: list[dict[str, str]] = attr.field(default=[])  # type: ignore
    message_prompt: str = attr.field(default="")
    message_edit: str = attr.field(default="")
    context: list[str] = attr.field(default=[])  # type: ignore
    diff_edit: str = attr.field(default="")
    test_command: str = attr.field(default="")
    version: str = attr.field(default=__version__)

    def save(self, fname: str | Path) -> None:
        with open(Path(fname), "w") as f:
            json.dump(attr.asdict(self), f, indent=4)

    @classmethod
    def load(cls, fname: str | Path) -> Sample:
        with open(fname, "r") as f:
            kwargs = json.load(f)
            _version = kwargs.get("version")
            if _version != __version__:
                raise SampleError(
                    f"Warning: sample version ({_version}) does not match current"
                    f" version ({__version__})."
                )
            return cls(**kwargs)



from textwrap import dedent

from mentat.code_feature import (
    CodeFeature,
    get_consolidated_feature_refs,
    split_file_into_intervals,
)
from mentat.interval import Interval


def test_split_file_into_intervals(temp_testbed, mock_session_context):
    with open("file_1.py", "w") as f:
        f.write(dedent("""\
            def func_1(x, y):
                return x + y
            
            def func_2():
                return 3
            """))
    code_feature = CodeFeature(mock_session_context.cwd / "file_1.py")
    interval_features = split_file_into_intervals(code_feature, 1)

    assert len(interval_features) == 2

    interval_1 = interval_features[0].interval
    interval_2 = interval_features[1].interval
    assert (interval_1.start, interval_1.end) == (1, 4)
    assert (interval_2.start, interval_2.end) == (4, 6)


def test_ref_method(temp_testbed):
    test_file = temp_testbed / "test_file.py"
    test_file.write_text("\n".join([""] * 10))
    code_feature = CodeFeature(test_file, Interval(2, 4))
    assert (code_feature.interval.start, code_feature.interval.end) == (2, 4)
    expected = temp_testbed / "test_file.py:2-4"
    assert str(code_feature) == str(expected)


def test_consolidated_refs(temp_testbed):
    scripts_dir = temp_testbed / "scripts"
    features = [
        CodeFeature(scripts_dir / "calculator.py", Interval(1, 10)),
        CodeFeature(scripts_dir / "calculator.py", Interval(10, 20)),
        CodeFeature(scripts_dir / "calculator.py", Interval(30, 40)),
        CodeFeature(scripts_dir / "echo.py"),
        CodeFeature(scripts_dir / "echo.py", Interval(10, 20)),
    ]
    consolidated = get_consolidated_feature_refs(features)
    assert len(consolidated) == 2
    assert consolidated[0] == f"{scripts_dir / 'calculator.py'}:1-10,10-20,30-40"
    assert consolidated[1] == str(scripts_dir / "echo.py")



from pathlib import Path
from textwrap import dedent
from typing import Any, AsyncGenerator, List

from mentat.llm_api_handler import chunk_to_lines
from mentat.parsers.file_edit import FileEdit, Replacement
from mentat.parsers.parser import ParsedLLMResponse
from mentat.session_context import SESSION_CONTEXT

# The git diff format should not be used with an LLM because it contains information like SHAs
# which it would not know about. It also involves certain arithmetic that it couldn't reliably do
# and typically prints out extra lines which aid human readability but would cost tokens.
# It is implemented to create training data from git diffs. Therefore there's not need for it to
# work asynchronously or stream partial results. In theory one could work this into the existing
# Parser class but it is simpler to assume one has the whole string up front and make heavy use of
# split.


class GitParser:
    # This doesn't actually "stream and parse" but it is named this way to match the interface of
    # the production parsers for use in the translation script.
    async def stream_and_parse_llm_response(
        self,
        response: AsyncGenerator[Any, None],
    ) -> ParsedLLMResponse:
        string = ""
        async for chunk in response:
            for content in chunk_to_lines(chunk):
                string += content
        return self.parse_string(string)

    def parse_string(self, git_diff: str) -> ParsedLLMResponse:
        session_context = SESSION_CONTEXT.get()

        # This is safe because actual code is prepended with ' ', + or -.
        split_on_diff = git_diff.split("\ndiff --git ")

        # Use commit message for conversation
        commit_message = ""
        if "\n\n" in split_on_diff[0]:
            commit_message = dedent(split_on_diff[0].split("\n\n")[1].strip())
            split_on_diff = split_on_diff[1:]
        else:
            split_on_diff[0] = split_on_diff[0].replace("diff --git ", "", 1)

        file_edits: List[FileEdit] = []
        for diff in split_on_diff:
            is_creation = "new file mode" in diff
            is_deletion = "deleted file mode" in diff
            first_line = diff.split("\n")[0]
            start_file_name = Path(first_line.split()[0][2:]).resolve()
            end_file_name = Path(first_line.split()[1][2:]).resolve()
            if start_file_name != end_file_name:
                new_name = end_file_name
            else:
                new_name = None

            file_edit = FileEdit(
                (session_context.cwd / start_file_name).resolve(),
                [],
                is_creation=is_creation,
                is_deletion=is_deletion,
                rename_file_path=new_name,
            )
            diff_split = diff.split("\n@@")
            if not is_deletion:
                for change in diff_split[1:]:
                    line_info = change.split("@@")[0]
                    # Git diff line represents line number information:
                    # @@ -a,b +c,d @@
                    # a is the original starting line number and b is the original number of lines.
                    # c and d are the new values.
                    # Both b and d are omitted when 1.
                    a_b = line_info.split()[0].split(",")
                    start_line = int(a_b[0][1:]) - 1
                    if len(a_b) == 1:
                        end_line = start_line + 1
                    else:
                        end_line = start_line + int(a_b[1])
                    line_changes = change.split("@@")[1]
                    code_lines = line_changes.split("\n")[
                        1:
                    ]  # Discard optional context
                    # This check is necessary because new code sometimes starts on the same line
                    # as @@ sometimes on the next line.
                    if code_lines[0] == "":
                        code_lines = code_lines[1:]
                    if code_lines[-1] == "":
                        code_lines = code_lines[:-1]

                    # Git diff gives context for human readability we don't want to train the llm
                    # to produce.
                    starting_repetition = 0
                    for line in code_lines:
                        if line.startswith(" "):
                            starting_repetition += 1
                        else:
                            break
                    ending_repetition = 0
                    for line in reversed(code_lines):
                        if line.startswith(" "):
                            ending_repetition += 1
                        else:
                            break
                    start_line += starting_repetition
                    end_line -= ending_repetition

                    lines: List[str] = []
                    for line in code_lines[
                        starting_repetition : len(code_lines) - ending_repetition
                    ]:
                        if not line.startswith("-"):
                            lines.append(line[1:])

                    file_edit.replacements.append(
                        Replacement(start_line, end_line, lines)
                    )

            file_edits.append(file_edit)

        return ParsedLLMResponse(git_diff, commit_message, file_edits)



import re
from pathlib import Path
from textwrap import dedent

import pytest
from git import Repo
from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionUserMessageParam,
)

from mentat.errors import SampleError
from mentat.git_handler import get_git_diff
from mentat.parsers.block_parser import BlockParser
from mentat.parsers.git_parser import GitParser
from mentat.python_client.client import PythonClient
from mentat.sampler import __version__
from mentat.sampler.sample import Sample
from mentat.sampler.sampler import Sampler
from mentat.sampler.utils import get_active_snapshot_commit
from mentat.session import Session
from scripts.sampler.run import run_sample


def remove_checksums(text):
    pattern = r"\b[0-9a-f]{7}\b"
    return re.sub(pattern, "", text)


@pytest.mark.asyncio
async def test_sample_from_context(
    mocker,
    temp_testbed,
    mock_session_context,
    mock_collect_user_input,
):
    mock_session_context.config.sample_repo = "test_sample_repo"

    mocker.patch(
        "mentat.conversation.Conversation.get_messages",
        return_value=[
            ChatCompletionSystemMessageParam(
                content="test_system_content",
                role="system",
            ),
            ChatCompletionUserMessageParam(
                content="test_user_content",
                role="user",
            ),
            ChatCompletionAssistantMessageParam(
                content="test_assistant_content",
                role="assistant",
            ),
        ],
    )

    mock_session_context.code_context.include(
        "multifile_calculator/operations.py",
    )

    with open("test_file.py", "w") as f:
        f.write("test_file_content\n")

    mock_collect_user_input.set_stream_messages(
        [
            "",
            "test_title",
            "test_description",
            "test_test_command",
        ]
    )
    sampler = Sampler()
    sample = await sampler.create_sample()
    assert sample.title == "test_title"
    assert sample.description == "test_description"
    assert sample.repo == "test_sample_repo"
    assert is_sha1(sample.merge_base)
    assert sample.diff_merge_base == ""
    assert sample.diff_active == ""
    assert sample.message_history == []
    assert sample.message_prompt == "test_user_content"
    assert sample.message_edit == "test_assistant_content"
    assert sample.context == ["multifile_calculator/operations.py"]
    expected_diff = (
        "diff --git a/test_file.py b/test_file.py\nnew file mode 100644\nindex"
        " 0000000..fffffff\n--- /dev/null\n+++ b/test_file.py\n@@ -0,0 +1"
        " @@\n+test_file_content\n"
    )
    assert remove_checksums(sample.diff_edit) == remove_checksums(expected_diff)
    assert sample.id != ""
    assert sample.test_command == "test_test_command"
    assert sample.version == __version__


def is_sha1(string: str) -> bool:
    return len(string) == 40 and all(c in "0123456789abcdef" for c in string)


@pytest.mark.asyncio
async def test_sample_command(temp_testbed, mock_collect_user_input, mock_call_llm_api):
    mock_collect_user_input.set_stream_messages(
        [
            "Request",
            "y",
            f"/sample {temp_testbed.as_posix()}",
            "",
            "test_url",
            "test_title",
            "test_description",
            "test_test_command",
            "q",
        ]
    )
    mock_call_llm_api.set_streamed_values([dedent("""\
        I will insert a comment in both files.

        @@start
        {
            "file": "multifile_calculator/calculator.py",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }
        @@code
        # forty two
        @@end
        @@start
        {
            "file": "test_file.py",
            "action": "create-file"
        }
        @@code
        # forty two
        @@end""")])

    session = Session(
        cwd=Path.cwd(), paths=[Path("multifile_calculator/calculator.py")]
    )
    session.start()
    await session.stream.recv(channel="client_exit")

    sample_files = list(temp_testbed.glob("sample_*.json"))
    assert len(sample_files) == 1
    sample = Sample.load(sample_files[0])

    assert sample.title == "test_title"
    assert sample.description == "test_description"
    assert sample.repo == "test_url"
    assert is_sha1(sample.merge_base)
    assert sample.diff_merge_base == ""
    assert sample.diff_active == ""
    assert sample.message_history == []
    assert sample.message_prompt == "Request"
    assert sample.message_edit == "I will insert a comment in both files."
    assert sample.context == [
        "multifile_calculator/calculator.py",
    ]
    edits = [e for e in sample.diff_edit.split("diff --git") if e]
    assert len(edits) == 2
    assert "multifile_calculator/calculator.py" in edits[0]
    assert "+# forty two" in edits[0]
    assert "test_file.py" in edits[1]
    assert "+# forty two" in edits[1]
    assert sample.test_command == "test_test_command"
    assert sample.version == "0.1.0"


test_sample = {
    "title": "Add sha1",
    "description": "",
    "id": "bc62d3903f4e4945a309ea0115d16702",
    "parent_id": "",
    "repo": "http://github.com/AbanteAI/mentat",
    "merge_base": "f5057f1658b9c7edb5e45a2fa8c2198ded5b5c00",
    "diff_merge_base": "",
    "diff_active": "",
    "message_history": [],
    "message_prompt": "Add a sha1 function to utils.py",
    "message_edit": (
        "I will add a new sha1 function to the `utils.py` file.\n\nSteps:\n1."
        " Add the sha1 function to `utils.py`."
    ),
    "context": ["mentat/utils.py"],
    "diff_edit": (
        "diff --git a/mentat/utils.py b/mentat/utils.py\nindex 46c3d7f..948b7f9"
        " 100644\n--- a/mentat/utils.py\n+++ b/mentat/utils.py\n@@ -35,2 +35,6 @@ def"
        " sha256(data: str) -> str:\n \n+def sha1(data: str) -> str:\n+    return"
        ' hashlib.sha1(data.encode("utf-8")).hexdigest()\n+\n+\n async def'
        " run_subprocess_async(*args: str) -> str:\n"
    ),
    "test_command": "",
    "version": "0.1.0",
}


@pytest.mark.asyncio
async def test_sample_eval(mock_call_llm_api):
    parsedLLMResponse = GitParser().parse_string(test_sample["diff_edit"])
    edit_message = BlockParser().file_edits_to_llm_message(parsedLLMResponse)
    mock_call_llm_api.set_streamed_values([dedent(f"""\
        I will add a new helper function called `sha1` to the `mentat/utils.py` file.
        
        Steps:
        1. Add the `sha1` function to `mentat/utils.py`.{edit_message}""")])

    sample = Sample(**test_sample)
    diff_eval = await run_sample(sample)
    assert remove_checksums(diff_eval) == remove_checksums(sample.diff_edit)


@pytest.mark.asyncio
async def test_sample_version_mismatch(temp_testbed):
    sample = Sample(**test_sample)
    sample.version = "0.0.9"
    sample_path = temp_testbed / "temp_sample.json"
    sample.save(sample_path)
    with pytest.raises(SampleError):
        Sample.load(sample_path)


def test_get_active_snapshot_commit(temp_testbed):
    repo = Repo(temp_testbed)
    # Add a test file and do an initial commit
    with open(temp_testbed / "test_file.py", "w") as f:
        f.write("test")
    repo.git.add("test_file.py")
    repo.git.commit("-m", "test commit")
    assert get_active_snapshot_commit(repo) is None  # No changes

    # Insert, Remove and Replace Lines
    with open(temp_testbed / "scripts" / "calculator.py", "r") as f:
        lines = f.readlines()
    lines = ["# Inserted Line\n"] + lines
    lines[20] = lines[20][:-2] + "# Replaced Line\n"
    lines = lines[:-4]  # "if __name__ == "__main__"
    with open(temp_testbed / "scripts" / "calculator.py", "w") as f:
        f.writelines(lines)

    # Create, Delete and Rename Files
    with open(temp_testbed / "scripts" / "calculator2.py", "w") as f:
        f.write("test")
    (temp_testbed / "scripts" / "echo.py").unlink()
    (temp_testbed / "scripts" / "graph_class.py").rename(
        temp_testbed / "scripts" / "graph.py"
    )

    commit_active = get_active_snapshot_commit(repo)

    # Confirm all changes in diff
    diff = repo.git.diff("HEAD", commit_active)
    assert "# Inserted Line" in diff
    assert "# Replaced Line" in diff
    assert "__main__" in diff
    assert "calculator2.py" in diff
    assert "echo.py" in diff
    assert "graph.py" in diff

    # Confirm current working files are unchanged
    with open(temp_testbed / "scripts" / "calculator.py", "r") as f:
        new_lines = f.readlines()
    for _line, _new_line in zip(lines, new_lines):
        assert _line == _new_line
    assert (temp_testbed / "scripts" / "calculator2.py").exists()
    assert not (temp_testbed / "scripts" / "echo.py").exists()
    assert (temp_testbed / "scripts" / "graph.py").exists()
    assert not (temp_testbed / "scripts" / "graph_class.py").exists()


def make_all_update_types(cwd, index):
    assert index in range(3)
    # Insert, Remove and Replace Lines
    with open(cwd / "multifile_calculator" / "operations.py", "r") as f:
        text_string = f.read()
    lines = text_string.split("\n")
    # Remove empty newline at end
    if lines[-1] == "":
        lines = lines[:-1]
    lines = [f"# Inserted Line {index}"] + lines
    lines[5] += f" # Replaced Line {index}"
    # Remove the last line and any empties before it
    lines = lines[:-1]
    while lines[-1] == "":
        lines = lines[:-1]
    # Add a final empty line
    lines.append("")
    with open(cwd / "multifile_calculator" / "operations.py", "w") as f:
        f.write("\n".join(lines))

    # Create, Delete and Rename Files
    with open(cwd / "multifile_calculator" / f"calculator{index}.py", "w") as f:
        f.write("test\n")
    format_examples = ["block.txt", "git_diff.txt", "replacement.txt"]
    (cwd / "format_examples" / format_examples[index]).unlink()
    old_name = "echo.py" if index == 0 else f"echo{index-1}.py"
    (cwd / "scripts" / old_name).rename(cwd / "scripts" / f"echo{index}.py")


def get_updates_as_parsed_llm_message(cwd):
    repo = Repo(cwd)
    starting_diff = repo.git.diff()
    # Commit active changes
    commit_active = get_active_snapshot_commit(repo)
    repo.git.add("--all")
    repo.git.commit("-m", "temporary commit")
    # Make diff_edit edits
    make_all_update_types(cwd, 2)
    diff_edit = get_git_diff("HEAD", cwd)
    parsedLLMResponse = GitParser().parse_string(diff_edit)
    # Reset hard and remove uncommitted files
    repo.git.reset("--hard")
    repo.git.clean("-fd")
    # Reset to previous commit, but keep changes as active
    repo.git.checkout("HEAD~1")
    # Re-apply the changes from commit_active without commiting
    repo.git.execute(["git", "cherry-pick", commit_active, "--no-commit"])
    # Unstage all changes
    repo.git.reset()
    ending_diff = repo.git.diff()
    if starting_diff != ending_diff:
        raise SampleError("Git state was not reset accurately.")

    return parsedLLMResponse


@pytest.mark.asyncio
async def test_sampler_integration(
    temp_testbed, mock_session_context, mock_call_llm_api
):
    # Setup the environemnt
    repo = Repo(temp_testbed)
    (temp_testbed / "test_file.py").write_text("permanent commit")
    repo.git.add("test_file.py")
    merge_base = repo.head.commit.hexsha
    # Make diff_merge_base edits + commit
    make_all_update_types(temp_testbed, 0)
    repo.git.add("--all")
    repo.git.commit("-m", "temporary commit")
    # Make diff_active edits
    make_all_update_types(temp_testbed, 1)

    # Verify it's setup correctly
    with open(temp_testbed / "multifile_calculator" / "operations.py", "r") as f:
        lines = f.readlines()
    for i in range(2):
        assert any(f"# Inserted Line {i}" in line for line in lines)
        assert any(f"# Replaced Line {i}" in line for line in lines)
    assert not any("# Inserted Line 2" in line for line in lines)
    assert not any("# Replaced Line 2" in line for line in lines)
    assert lines[-1] == "    return a - b\n"
    assert (temp_testbed / "multifile_calculator" / "calculator0.py").exists()
    assert (temp_testbed / "multifile_calculator" / "calculator1.py").exists()
    assert not (temp_testbed / "format_examples" / "block.txt").exists()
    assert not (temp_testbed / "format_examples" / "git_diff.txt").exists()
    assert (temp_testbed / "format_examples" / "replacement.txt").exists()

    # Generate file_edits to be performed for test, verify they're setup correctly
    parsed_llm_message = get_updates_as_parsed_llm_message(temp_testbed)
    file_edits = parsed_llm_message.file_edits
    assert any("Inserted Line 2" in str(f.replacements) for f in file_edits)
    assert any("Replaced Line 2" in str(f.replacements) for f in file_edits)
    assert any(
        # In file-edit, the entire file is overwritten, so we verify it's missing the last line
        (
            "operations.py" in str(f.file_path)
            and "return a - b" not in str(f.replacements)
        )
        for f in file_edits
    )
    assert any("calculator2.py" in str(f.file_path) for f in file_edits)
    assert any("replacement.txt" in str(f.file_path) for f in file_edits)

    llm_response = BlockParser().file_edits_to_llm_message(parsed_llm_message)
    mock_call_llm_api.set_streamed_values(
        [f"I will make the following edits. {llm_response}"]
    )

    # Generate a sample using Mentat
    python_client = PythonClient(cwd=temp_testbed, paths=["."])
    await python_client.startup()
    await python_client.call_mentat_auto_accept(dedent("""\
        Make the following changes to "multifile_calculator/operations.py":
        1. Add "# Inserted line 2" as the first line
        2. Add "# Replaced Line 2" to the end of the 5th line
        3. Remove the last non-blank line
        Make the following other changes:
        4. Create "multifile_calculator/calculator2.py"
        5. Delete "format_examples/replacement.txt"
        6. Rename "scripts/echo1.py" to "scripts/echo2.py"
        """))

    # Remove all included files; rely on the diff to include them
    python_client.session.ctx.code_context.include_files = {}
    await python_client.call_mentat(f"/sample {temp_testbed.as_posix()}")
    await python_client.call_mentat(merge_base)
    await python_client.call_mentat("test_url")
    await python_client.call_mentat("test_title")
    await python_client.call_mentat("test_description")
    await python_client.call_mentat("test_test_command")
    await python_client.call_mentat("q")
    await python_client.shutdown()

    # Evaluate the sample using Mentat
    sample_files = list(temp_testbed.glob("sample_*.json"))
    assert len(sample_files) == 1
    sample = Sample.load(sample_files[0])
    assert sample.title == "test_title"
    assert sample.description == "test_description"
    assert sample.repo == "test_url"
    assert sample.merge_base == merge_base
    assert sample.diff_merge_base != ""
    assert sample.diff_active != ""
    assert sample.message_history == []
    assert sample.message_edit == "I will make the following edits. "
    assert set(sample.context) == {
        "scripts/echo1.py",
        "multifile_calculator/operations.py",
        "format_examples/replacement.txt",
    }
    assert sample.diff_edit != ""

    mock_call_llm_api.set_streamed_values(
        [f"I will make the following edits. {llm_response}"]
    )
    diff_eval = await run_sample(sample, temp_testbed)
    assert diff_eval == sample.diff_edit



import os
import subprocess

from mentat.git_handler import get_git_diff, get_hexsha_active


def test_get_git_diff(temp_testbed, mock_session_context):
    # Add a new file and commit it, then delete but don't commit
    with open(temp_testbed / "test_file.txt", "w") as f:
        f.write("forty two")
    assert "forty two" in get_git_diff("HEAD")
    subprocess.run(["git", "add", "."])
    subprocess.run(["git", "commit", "-m", "test commit"])
    assert "forty two" not in get_git_diff("HEAD")
    assert "forty two" in get_git_diff("HEAD~1", "HEAD")
    os.remove("test_file.txt")
    assert "forty two" in get_git_diff("HEAD")
    assert "forty two" not in get_git_diff("HEAD~1")


def test_get_hexsha_active(temp_testbed):
    a = get_hexsha_active()
    with open("multifile_calculator/calculator.py", "a") as f:
        f.write("forty three")
    b = get_hexsha_active()
    with open("test_file.txt", "w") as f:
        f.write("forty two")
    c = get_hexsha_active()
    assert a != b
    assert b != c
    assert a != c



import sys
from unittest.mock import AsyncMock

import pytest

from mentat.terminal.client import TerminalClient

# These ui-tests won't run automatically
# To run them, use pytest -s --uitest
# Without the -s, pytest will capture all
# i/o and the tests will not work
pytestmark = pytest.mark.uitest


def is_test_correct():
    ans = input("Does this look correct? (Y/n)").lower()
    sys.exit(0 if ans in ["", "y"] else 1)


@pytest.fixture
def ui_mock_collect_user_input(
    mocker,
):
    async_mock = AsyncMock()

    mocker.patch("mentat.code_edit_feedback.collect_user_input", side_effect=async_mock)
    mocker.patch("mentat.session_input.collect_user_input", side_effect=async_mock)

    async_mock.side_effect = is_test_correct
    return async_mock


def test_start(mock_call_llm_api, ui_mock_collect_user_input):
    print()
    with pytest.raises(SystemExit) as e_info:
        terminal_client = TerminalClient()
        terminal_client.run()
    assert e_info.value.code == 0



import subprocess
from pathlib import Path
from textwrap import dedent

import pytest

from mentat.code_feature import CodeFeature
from mentat.command.command import Command, InvalidCommand
from mentat.command.commands.context import ContextCommand
from mentat.command.commands.help import HelpCommand
from mentat.session import Session
from mentat.session_context import SESSION_CONTEXT


def test_invalid_command():
    assert isinstance(Command.create_command("non-existent"), InvalidCommand)


@pytest.mark.asyncio
async def test_help_command(mock_call_llm_api):
    command = Command.create_command("help")
    await command.apply()
    assert isinstance(command, HelpCommand)


@pytest.mark.asyncio
async def test_commit_command(temp_testbed, mock_collect_user_input):
    file_name = "test_file.py"
    with open(file_name, "w") as f:
        f.write("# Commit me!")

    mock_collect_user_input.set_stream_messages(
        [
            "/commit",
            "q",
        ]
    )

    session = Session(cwd=temp_testbed, paths=[])
    session.start()
    await session.stream.recv(channel="client_exit")

    assert subprocess.check_output(["git", "status", "-s"], text=True) == ""


# TODO: test without git
@pytest.mark.asyncio
async def test_include_command(temp_testbed, mock_collect_user_input):
    mock_collect_user_input.set_stream_messages(
        [
            "/include scripts",
            "q",
        ]
    )

    session = Session(cwd=temp_testbed)
    session.start()
    await session.stream.recv(channel="client_exit")

    code_context = SESSION_CONTEXT.get().code_context
    assert (
        Path(temp_testbed) / "scripts" / "calculator.py" in code_context.include_files
    )


# TODO: test without git
@pytest.mark.asyncio
async def test_exclude_command(temp_testbed, mock_collect_user_input):
    mock_collect_user_input.set_stream_messages(
        [
            "/exclude scripts",
            "q",
        ]
    )

    session = Session(cwd=temp_testbed, paths=["scripts"])
    session.start()
    await session.stream.recv(channel="client_exit")

    code_context = SESSION_CONTEXT.get().code_context
    assert not code_context.include_files


@pytest.mark.asyncio
async def test_undo_command(temp_testbed, mock_collect_user_input, mock_call_llm_api):
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
            # This is a temporary file
            # with 2 lines"""))

    mock_collect_user_input.set_stream_messages(
        [
            "Edit the file",
            "y",
            "/undo",
            "q",
        ]
    )

    mock_call_llm_api.set_streamed_values([dedent(f"""\
        Conversation

        @@start
        {{
            "file": "{temp_file_name}",
            "action": "insert",
            "insert-after-line": 1,
            "insert-before-line": 2
        }}
        @@code
        # I inserted this comment
        @@end""")])

    session = Session(cwd=temp_testbed, paths=[temp_file_name])
    session.start()
    await session.stream.recv(channel="client_exit")

    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
            # This is a temporary file
            # with 2 lines""")
    assert content == expected_content


@pytest.mark.asyncio
async def test_redo_command(temp_testbed, mock_collect_user_input, mock_call_llm_api):
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
            # This is a temporary file
            # with 2 lines"""))

    mock_collect_user_input.set_stream_messages(
        [
            "Edit the file",
            "y",
            "/undo",
            "/redo",
            "q",
        ]
    )

    new_file_name = "new_temp.py"
    mock_call_llm_api.set_streamed_values([dedent(f"""\
        Conversation

        @@start
        {{
            "file": "{temp_file_name}",
            "action": "insert",
            "insert-after-line": 1,
            "insert-before-line": 2
        }}
        @@code
        # I inserted this comment
        @@end
        @@start
        {{
            "file": "{new_file_name}",
            "action": "create-file"
        }}
        @@code
        # I created this file
        @@end
        """)])

    session = Session(cwd=Path.cwd(), paths=[temp_file_name])
    session.start()
    await session.stream.recv(channel="client_exit")

    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
            # This is a temporary file
            # I inserted this comment
            # with 2 lines""")
    assert content == expected_content

    with open(new_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
            # I created this file""")
    assert content == expected_content


@pytest.mark.asyncio
async def test_undo_all_command(
    temp_testbed, mock_collect_user_input, mock_call_llm_api
):
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
            # This is a temporary file
            # with 2 lines"""))

    mock_collect_user_input.set_stream_messages(
        [
            "",
            "y",
            "/undo-all",
            "q",
        ]
    )

    # TODO: Make a way to set multiple return values for call_llm_api and reset multiple edits at once
    mock_call_llm_api.set_streamed_values([dedent(f"""\
        Conversation

        @@start
        {{
            "file": "{temp_file_name}",
            "action": "insert",
            "insert-after-line": 1,
            "insert-before-line": 2
        }}
        @@code
        # I inserted this comment
        @@end""")])

    session = Session(cwd=temp_testbed, paths=[temp_file_name])
    session.start()
    await session.stream.recv(channel="client_exit")

    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
            # This is a temporary file
            # with 2 lines""")
    assert content == expected_content


@pytest.mark.asyncio
async def test_clear_command(temp_testbed, mock_collect_user_input, mock_call_llm_api):
    mock_collect_user_input.set_stream_messages(
        [
            "Request",
            "/clear",
            "q",
        ]
    )
    mock_call_llm_api.set_streamed_values(["Answer"])

    session = Session(cwd=Path.cwd())
    session.start()
    await session.stream.recv(channel="client_exit")

    conversation = SESSION_CONTEXT.get().conversation
    assert len(conversation.get_messages()) == 1


# TODO: test without git
@pytest.mark.asyncio
async def test_search_command(
    mocker, temp_testbed, mock_call_llm_api, mock_collect_user_input
):
    mock_collect_user_input.set_stream_messages(
        [
            "Request",
            "/search Query",
            "q",
        ]
    )
    mock_call_llm_api.set_streamed_values(["Answer"])
    mock_feature = CodeFeature(
        Path(temp_testbed) / "multifile_calculator" / "calculator.py"
    )
    mock_score = 1.0
    mocker.patch(
        "mentat.code_context.CodeContext.search",
        return_value=[(mock_feature, mock_score)],
    )
    session = Session(cwd=Path.cwd())
    session.start()
    await session.stream.recv(channel="client_exit")

    rel_path = mock_feature.path.relative_to(Path(temp_testbed))
    assert str(rel_path) in session.stream.messages[-3].data
    assert "cost" in session.stream.messages[-2].data


@pytest.mark.asyncio
async def test_context_command(temp_testbed, mock_call_llm_api):
    command = Command.create_command("context")
    await command.apply()
    assert isinstance(command, ContextCommand)


@pytest.mark.asyncio
async def test_config_command(mock_call_llm_api):
    session_context = SESSION_CONTEXT.get()
    config = session_context.config
    stream = session_context.stream
    command = Command.create_command("config")
    await command.apply("test")
    assert stream.messages[-1].data == "Unrecognized config option: test"
    await command.apply("model")
    assert stream.messages[-1].data.startswith("model: ")
    await command.apply("model", "test")
    assert stream.messages[-1].data == "model set to test"
    assert config.model == "test"
    await command.apply("model", "test", "lol")
    assert stream.messages[-1].data == "Too many arguments"


@pytest.mark.asyncio
async def test_screenshot_command(mocker):
    # Mock the session context and its attributes
    session_context = SESSION_CONTEXT.get()
    mock_vision_manager = mocker.MagicMock()
    session_context.vision_manager = mock_vision_manager
    config = session_context.config
    stream = session_context.stream
    conversation = session_context.conversation

    assert config.model != "gpt-4-vision-preview"

    mock_vision_manager.screenshot.return_value = "fake_image_data"

    screenshot_command = Command.create_command("screenshot")
    await screenshot_command.apply("fake_path")

    mock_vision_manager.screenshot.assert_called_once_with("fake_path")
    assert config.model == "gpt-4-vision-preview"
    assert stream.messages[-1].data == "Screenshot taken for: fake_path."
    assert conversation._messages[-1] == {
        "role": "user",
        "content": [
            {"type": "text", "text": "A screenshot of fake_path"},
            {"type": "image_url", "image_url": {"url": "fake_image_data"}},
        ],
    }

    # Test non-gpt models aren't changed
    config.model = "test"
    await screenshot_command.apply("fake_path")
    assert config.model == "test"

    # Test other models containing vision aren't changed
    config.model = "gpt-vision"
    await screenshot_command.apply("fake_path")
    assert config.model == "gpt-vision"



from pathlib import Path

import pytest

from mentat.code_feature import CodeFeature
from mentat.embeddings import get_feature_similarity_scores


def _make_code_feature(path, text):
    with open(path, "w") as f:
        f.write(text)
    return CodeFeature(path)


@pytest.mark.asyncio
async def test_get_feature_similarity_scores(mocker, mock_call_embedding_api):
    prompt = "example prompt"
    features = [
        _make_code_feature(Path(f"file{i}.txt").resolve(), f"File {i}")
        for i in range(3)
    ]
    mock_call_embedding_api.set_embedding_values(
        [
            [0.7, 0.7, 0.7],  # The prompt
            [0.4, 0.4, 0.4],
            [0.5, 0.6, 0.7],
            [0.69, 0.7, 0.71],
        ]
    )
    result = await get_feature_similarity_scores(prompt, features)
    assert len(result) == 3
    assert max(result) == result[0]  # The first feature is most similar



from enum import Enum
from pathlib import Path

import attr
from pygments import highlight  # pyright: ignore[reportUnknownVariableType]
from pygments.formatters import TerminalFormatter
from pygments.lexer import Lexer
from pygments.lexers import TextLexer, get_lexer_for_filename
from pygments.util import ClassNotFound
from termcolor import colored

from mentat.session_context import SESSION_CONTEXT
from mentat.utils import get_relative_path

change_delimiter = 60 * "="


def _get_lexer(file_path: Path):
    try:
        lexer: Lexer = get_lexer_for_filename(file_path)
    except ClassNotFound:
        lexer = TextLexer()
    lexer.stripnl = False
    lexer.stripall = False
    lexer.ensurenl = False
    return lexer


def get_line_number_buffer(file_lines: list[str]):
    return len(str(len(file_lines) + 1)) + 1


class FileActionType(Enum):
    RenameFile = "rename"
    CreateFile = "create"
    DeleteFile = "delete"
    UpdateFile = "update"


def get_file_action_type(is_creation: bool, is_deletion: bool, new_name: Path | None):
    if is_creation:
        file_action_type = FileActionType.CreateFile
    elif is_deletion:
        file_action_type = FileActionType.DeleteFile
    elif new_name is not None:
        file_action_type = FileActionType.RenameFile
    else:
        file_action_type = FileActionType.UpdateFile
    return file_action_type


@attr.define(slots=False)
class DisplayInformation:
    file_name: Path = attr.field()
    file_lines: list[str] = attr.field()
    added_block: list[str] = attr.field()
    removed_block: list[str] = attr.field()
    file_action_type: FileActionType = attr.field()
    first_changed_line: int = attr.field(default=0)
    last_changed_line: int = attr.field(default=0)
    new_name: Path | None = attr.field(default=None)

    def __attrs_post_init__(self):
        ctx = SESSION_CONTEXT.get()

        self.line_number_buffer = get_line_number_buffer(self.file_lines)
        self.lexer = _get_lexer(self.file_name)

        if self.file_name.is_absolute():
            self.file_name = get_relative_path(self.file_name, ctx.cwd)
        if self.new_name is not None and self.new_name.is_absolute():
            self.new_name = get_relative_path(self.new_name, ctx.cwd)


def _remove_extra_empty_lines(lines: list[str]) -> list[str]:
    if not lines:
        return []

    # Find the first non-empty line
    start = 0
    while start < len(lines) and not lines[start].strip():
        start += 1

    # Find the last non-empty line
    end = len(lines) - 1
    while end > start and not lines[end].strip():
        end -= 1

    # If all lines are empty, keep only one empty line
    if start == len(lines):
        return [" "]

    # Return the list with only a maximum of one empty line on either side
    return lines[max(start - 1, 0) : end + 2]


def _prefixed_lines(line_number_buffer: int, lines: list[str], prefix: str):
    return "\n".join(
        [
            prefix + " " * (line_number_buffer - len(prefix)) + line.strip("\n")
            for line in lines
        ]
    )


def _get_code_block(
    code_lines: list[str],
    line_number_buffer: int,
    prefix: str,
    color: str | None,
):
    lines = _prefixed_lines(line_number_buffer, code_lines, prefix)
    if lines:
        return "\n".join(colored(line, color=color) for line in lines.split("\n"))
    else:
        return ""


def get_full_change(display_information: DisplayInformation, prefix: str = ""):
    to_print = [
        get_file_name(display_information),
        (
            change_delimiter
            if display_information.added_block or display_information.removed_block
            else ""
        ),
        get_previous_lines(display_information),
        get_removed_lines(display_information),
        get_added_lines(display_information),
        get_later_lines(display_information),
        (
            change_delimiter
            if display_information.added_block or display_information.removed_block
            else ""
        ),
    ]
    full_change = "\n".join([line for line in to_print if line])
    prefixed_change = "\n".join(
        (prefix + line) if line.strip() else line for line in full_change.split("\n")
    )
    return prefixed_change


def get_file_name(
    display_information: DisplayInformation,
):
    match display_information.file_action_type:
        case FileActionType.CreateFile:
            return "\n" + colored(
                f"{display_information.file_name}*", color="light_green"
            )
        case FileActionType.DeleteFile:
            return "\n" + colored(
                f"Deletion: {display_information.file_name}", color="light_red"
            )
        case FileActionType.RenameFile:
            return "\n" + colored(
                f"Rename: {display_information.file_name} ->"
                f" {display_information.new_name}",
                color="yellow",
            )
        case FileActionType.UpdateFile:
            return "\n" + colored(
                f"{display_information.file_name}", color="light_blue"
            )


def get_added_lines(
    display_information: DisplayInformation,
    prefix: str = "+",
    color: str | None = "green",
):
    return _get_code_block(
        display_information.added_block,
        display_information.line_number_buffer,
        prefix,
        color,
    )


def get_removed_lines(
    display_information: DisplayInformation,
    prefix: str = "-",
    color: str | None = "red",
):
    return _get_code_block(
        display_information.removed_block,
        display_information.line_number_buffer,
        prefix,
        color,
    )


def highlight_text(display_information: DisplayInformation, text: str) -> str:
    # pygments doesn't have type hints on TerminalFormatter
    return highlight(text, display_information.lexer, TerminalFormatter(bg="dark"))  # type: ignore


def get_previous_lines(
    display_information: DisplayInformation,
    num: int = 2,
) -> str:
    if display_information.first_changed_line < 0:
        return ""
    lines = _remove_extra_empty_lines(
        [
            display_information.file_lines[i]
            for i in range(
                max(0, display_information.first_changed_line - num),
                min(
                    display_information.first_changed_line,
                    len(display_information.file_lines),
                ),
            )
        ]
    )
    numbered = [
        (str(display_information.first_changed_line - len(lines) + i + 1) + ":").ljust(
            display_information.line_number_buffer
        )
        + line
        for i, line in enumerate(lines)
    ]

    prev = "\n".join(numbered)
    return highlight_text(display_information, prev)


def get_later_lines(
    display_information: DisplayInformation,
    num: int = 2,
) -> str:
    if display_information.last_changed_line < 0:
        return ""
    lines = _remove_extra_empty_lines(
        [
            display_information.file_lines[i]
            for i in range(
                max(0, display_information.last_changed_line),
                min(
                    display_information.last_changed_line + num,
                    len(display_information.file_lines),
                ),
            )
        ]
    )
    numbered = [
        (str(display_information.last_changed_line + 1 + i) + ":").ljust(
            display_information.line_number_buffer
        )
        + line
        for i, line in enumerate(lines)
    ]

    later = "\n".join(numbered)
    return highlight_text(display_information, later)



import asyncio

import pytest

from mentat.session_context import SESSION_CONTEXT
from mentat.session_input import ask_yes_no


@pytest.mark.asyncio
async def test_ask_yes_no():
    stream = SESSION_CONTEXT.get().stream
    stream.start()

    # Test when user inputs invalid confirmation
    ask_yes_no_task = asyncio.create_task(ask_yes_no(default_yes=False))
    input_request_message = await stream.recv("input_request")
    stream.send("yes", channel=f"input_request:{input_request_message.id}")
    input_request_message = await stream.recv("input_request")
    stream.send("y", channel=f"input_request:{input_request_message.id}")
    assert await ask_yes_no_task is True

    # Test when user inputs 'Y'
    ask_yes_no_task = asyncio.create_task(ask_yes_no(default_yes=False))
    input_request_message = await stream.recv("input_request")
    stream.send("Y", channel=f"input_request:{input_request_message.id}")
    assert await ask_yes_no_task is True

    # Test when user inputs nothing
    ask_yes_no_task = asyncio.create_task(ask_yes_no(default_yes=False))
    input_request_message = await stream.recv("input_request")
    stream.send("", channel=f"input_request:{input_request_message.id}")
    assert await ask_yes_no_task is False

    # Test when user inputs 'n'
    ask_yes_no_task = asyncio.create_task(ask_yes_no(default_yes=False))
    input_request_message = await stream.recv("input_request")
    stream.send("n", channel=f"input_request:{input_request_message.id}")
    assert await ask_yes_no_task is False

    # Test default true
    ask_yes_no_task = asyncio.create_task(ask_yes_no(default_yes=True))
    input_request_message = await stream.recv("input_request")
    stream.send("", channel=f"input_request:{input_request_message.id}")
    assert await ask_yes_no_task is True

    stream.stop()



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, AsyncGenerator, Dict, List, cast
from uuid import UUID, uuid4

from mentat.broadcast import Broadcast


class StreamMessageSource(Enum):
    SERVER = "server"
    CLIENT = "client"


@dataclass(slots=True)
class StreamMessage:
    id: UUID
    channel: str
    source: StreamMessageSource
    data: Any
    extra: Dict[str, Any]
    created_at: datetime


class SessionStream:
    """Replaces `cprint` and `print`

    Stores message history for a Session and holds an in-memory message bus.

    Terminal and extension clients can read these messages and render them accordingly.
    For the terminal, they would be rendered with `cprint`.
    """

    def __init__(self):
        self.messages: List[StreamMessage] = []
        self.interrupt_lock = asyncio.Lock()
        self._broadcast = Broadcast()

    def start(self):
        self._broadcast.connect()

    def stop(self):
        self._broadcast.disconnect()

    # Since there is no maximum queue size, use the synchronous version of this function instead
    async def send_async(
        self,
        data: Any,
        source: StreamMessageSource = StreamMessageSource.SERVER,
        channel: str = "default",
        **kwargs: Any,
    ):
        message = StreamMessage(
            id=uuid4(),
            source=source,
            channel=channel,
            data=data,
            created_at=datetime.utcnow(),
            extra=kwargs,
        )

        self.messages.append(message)
        await self._broadcast.publish_async(channel=channel, message=message)

        return message

    def send(
        self,
        data: Any,
        source: StreamMessageSource = StreamMessageSource.SERVER,
        channel: str = "default",
        **kwargs: Any,
    ):
        message = StreamMessage(
            id=uuid4(),
            source=source,
            channel=channel,
            data=data,
            created_at=datetime.utcnow(),
            extra=kwargs,
        )

        self.messages.append(message)
        self._broadcast.publish(channel=channel, message=message)

        return message

    async def recv(self, channel: str = "default") -> StreamMessage:
        """Listen for a single event on a channel"""
        with self._broadcast.subscribe(channel) as subscriber:
            async for event in subscriber:
                stream_message = cast(StreamMessage, event.message)
                return stream_message
            raise Exception("recv should not complete without receiving an Event")

    async def listen(
        self, channel: str = "default"
    ) -> AsyncGenerator[StreamMessage, None]:
        """Listen to all messages on a channel indefinitely"""
        with self._broadcast.subscribe(channel) as subscriber:
            async for event in subscriber:
                yield event.message

    async def join(self) -> None:
        """Blocks until all sent events have been processed"""
        await self._broadcast.join()



import base64
from io import BytesIO

from PIL import Image

from mentat.llm_api_handler import prompt_tokens


def test_prompt_tokens():
    messages = [
        {"role": "user", "content": "Hello!"},
        {"role": "assistant", "content": "Hi there! How can I help you today?"},
    ]
    model = "gpt-4-vision-preview"

    assert prompt_tokens(messages, model) == 24

    # An image that must be scaled twice and then fits in 6 512x512 panels
    img = Image.new("RGB", (768 * 4, 1050 * 4), color="red")
    buffer = BytesIO()
    img.save(buffer, format="PNG")
    buffer.seek(0)
    img_base64 = base64.b64encode(buffer.getvalue()).decode()
    image_url = f"data:image/png;base64,{img_base64}"

    messages.append(
        {
            "role": "user",
            "content": [{"type": "image_url", "image_url": {"url": image_url}}],
        }
    )

    assert prompt_tokens(messages, model) == 24 + 6 * 170 + 85 + 5



import pytest

from mentat.broadcast import Broadcast


@pytest.mark.asyncio
async def test_broadcast():
    with Broadcast() as broadcast:
        with broadcast.subscribe("chatroom") as subscriber:
            broadcast.publish("chatroom", "hello")
            event = await subscriber.get()
            assert event.channel == "chatroom"
            assert event.message == "hello"



import json
import os
import sqlite3
from pathlib import Path
from timeit import default_timer

import numpy as np

from mentat.code_feature import CodeFeature, count_feature_tokens
from mentat.errors import MentatError
from mentat.llm_api_handler import (
    count_tokens,
    model_context_size,
    model_price_per_1000_tokens,
)
from mentat.session_context import SESSION_CONTEXT
from mentat.session_input import ask_yes_no
from mentat.utils import mentat_dir_path, sha256

EMBEDDINGS_API_BATCH_SIZE = 2048


class EmbeddingsDatabase:
    def __init__(self, output_dir: Path | None = None):
        self.output_dir = output_dir or mentat_dir_path
        os.makedirs(self.output_dir, exist_ok=True)
        self.path = Path(self.output_dir) / "embeddings.sqlite3"
        self._connect()

    def _connect(self):
        self.conn = sqlite3.connect(self.path)
        with self.conn as db:
            db.execute(
                "CREATE TABLE IF NOT EXISTS embeddings "
                "(checksum TEXT PRIMARY KEY, vector BLOB)"
            )

    def set(self, items: dict[str, list[float]]):
        with self.conn as db:
            db.executemany(
                "INSERT OR REPLACE INTO embeddings (checksum, vector) VALUES (?, ?)",
                [
                    (key, sqlite3.Binary(json.dumps(value).encode("utf-8")))
                    for key, value in items.items()
                ],
            )

    def get(self, keys: list[str]) -> dict[str, list[float]]:
        with self.conn as db:
            cursor = db.execute(
                "SELECT checksum, vector FROM embeddings WHERE checksum IN"
                f" ({','.join(['?']*len(keys))})",
                keys,
            )
            return {row[0]: json.loads(row[1]) for row in cursor.fetchall()}

    def exists(self, key: str) -> bool:
        with self.conn as db:
            cursor = db.execute("SELECT 1 FROM embeddings WHERE checksum=?", (key,))
            return cursor.fetchone() is not None

    def __del__(self):
        self.conn.close()


database = EmbeddingsDatabase()


def _cosine_similarity(v1: list[float], v2: list[float]) -> float:
    """Calculate the cosine similarity between two vectors."""
    dot_product = np.dot(v1, v2)
    norm_v1 = np.linalg.norm(v1)
    norm_v2 = np.linalg.norm(v2)
    return dot_product / (norm_v1 * norm_v2)  # pyright: ignore


async def get_feature_similarity_scores(
    prompt: str,
    features: list[CodeFeature],
    loading_multiplier: float = 0.0,
) -> list[float]:
    """Return the similarity scores for a given prompt and list of features."""
    global database
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream
    cost_tracker = session_context.cost_tracker
    embedding_model = session_context.config.embedding_model
    llm_api_handler = session_context.llm_api_handler

    max_model_tokens = model_context_size(embedding_model)
    if max_model_tokens is None:
        raise MentatError(f"Missing model context size for {embedding_model}.")
    prompt_tokens = count_tokens(prompt, embedding_model, False)
    if prompt_tokens > max_model_tokens:
        stream.send(
            f"Warning: Prompt contains {prompt_tokens} tokens, but the model"
            f" can only handle {max_model_tokens} tokens. Ignoring embeddings."
        )
        return [0.0 for _ in features]

    prompt_checksum = sha256(prompt)
    checksums: list[str] = [f.get_checksum() for f in features]
    tokens: list[int] = await count_feature_tokens(features, embedding_model)
    embed_texts = list[str]()
    embed_checksums = list[str]()
    embed_tokens = list[int]()
    if not database.exists(prompt_checksum):
        embed_texts.append(prompt)
        embed_checksums.append(prompt_checksum)
        embed_tokens.append(prompt_tokens)
    for feature, checksum, token in zip(features, checksums, tokens):
        if token > max_model_tokens:
            continue
        if not database.exists(checksum):
            embed_texts.append("\n".join(feature.get_code_message()))
            embed_checksums.append(checksum)
            embed_tokens.append(token)

    # If it costs more than $1, get confirmation from user.
    cost = model_price_per_1000_tokens(embedding_model)
    if cost is None:
        stream.send(
            "Warning: Could not determine cost of embeddings. Continuing anyway.",
            style="warning",
        )
    else:
        expected_cost = (sum(embed_tokens) / 1000) * cost[0]
        if expected_cost > 1.0:
            stream.send(
                f"Embedding {sum(embed_tokens)} tokens will cost ${cost[0]:.2f}."
                " Continue anyway?"
            )
            if not await ask_yes_no(default_yes=True):
                stream.send("Ignoring embeddings for now.")
                return [0.0 for _ in checksums]

    # Fetch embeddings in batches
    if len(embed_texts) == 0:
        n_batches = 0
    else:
        n_batches = len(embed_texts) // EMBEDDINGS_API_BATCH_SIZE + 1
    for batch in range(n_batches):
        if loading_multiplier:
            stream.send(
                f"Fetching embeddings, batch {batch+1}/{n_batches}",
                channel="loading",
                progress=(100 / n_batches) * loading_multiplier,
            )
        start_time = default_timer()
        i_start, i_end = (
            batch * EMBEDDINGS_API_BATCH_SIZE,
            (batch + 1) * EMBEDDINGS_API_BATCH_SIZE,
        )
        _texts = embed_texts[i_start:i_end]
        _checksums = embed_checksums[i_start:i_end]
        _tokens = embed_tokens[i_start:i_end]

        response = await llm_api_handler.call_embedding_api(_texts, embedding_model)
        cost_tracker.log_api_call_stats(
            sum(_tokens),
            0,
            embedding_model,
            start_time - default_timer(),
        )
        database.set({k: v for k, v in zip(_checksums, response)})

    # Calculate similarity score for each feature
    prompt_embedding = database.get([prompt_checksum])[prompt_checksum]
    embeddings = database.get(checksums)
    scores = [
        _cosine_similarity(prompt_embedding, embeddings[k]) if k in embeddings else 0.0
        for k in checksums
    ]

    return scores



from mentat.parsers.block_parser import BlockParser
from mentat.parsers.replacement_parser import ReplacementParser
from mentat.session_context import SESSION_CONTEXT


def test_midconveration_parser_change(mock_call_llm_api):
    session_context = SESSION_CONTEXT.get()
    config = session_context.config
    conversation = session_context.conversation

    config.parser = "block"
    assert (
        conversation.get_messages()[0]["content"] == BlockParser().get_system_prompt()
    )

    config.parser = "replacement"
    assert (
        conversation.get_messages()[0]["content"]
        == ReplacementParser().get_system_prompt()
    )


def test_no_parser_prompt(mock_call_llm_api):
    session_context = SESSION_CONTEXT.get()
    config = session_context.config
    conversation = session_context.conversation

    assert len(conversation.get_messages()) == 1
    config.no_parser_prompt = True
    assert len(conversation.get_messages()) == 0


def test_add_user_message_with_and_without_image(mock_call_llm_api):
    session_context = SESSION_CONTEXT.get()
    conversation = session_context.conversation

    # Test with image
    test_message = "Hello, World!"
    test_image_url = "http://example.com/image.png"
    conversation.add_user_message(test_message, test_image_url)
    messages_with_image = conversation.get_messages()
    assert len(messages_with_image) == 2  # System prompt + user message
    user_message_content_with_image = messages_with_image[-1]["content"]
    assert len(user_message_content_with_image) == 2  # Text + image
    assert user_message_content_with_image[0]["type"] == "text"
    assert user_message_content_with_image[0]["text"] == test_message
    assert user_message_content_with_image[1]["type"] == "image_url"
    assert user_message_content_with_image[1]["image_url"]["url"] == test_image_url

    # Test without image
    conversation.clear_messages()
    conversation.add_user_message(test_message)
    messages_without_image = conversation.get_messages()
    assert len(messages_without_image) == 2  # System prompt + user message
    user_message_content_without_image = messages_without_image[-1]["content"]
    assert user_message_content_without_image == test_message



from typing import List

from typing_extensions import override

from mentat.command.command import Command, CommandArgument
from mentat.session_context import SESSION_CONTEXT


class ContextCommand(Command, command_name="context"):
    @override
    async def apply(self, *args: str) -> None:
        session_context = SESSION_CONTEXT.get()
        code_context = session_context.code_context

        code_context.display_context()

    @override
    @classmethod
    def arguments(cls) -> List[CommandArgument]:
        return []

    @override
    @classmethod
    def argument_autocompletions(
        cls, arguments: list[str], argument_position: int
    ) -> list[str]:
        return []

    @override
    @classmethod
    def help_message(cls) -> str:
        return "Show all files currently in context."



import os
from pathlib import Path
from textwrap import dedent

import pytest

from mentat.parsers.file_edit import FileEdit, Replacement
from mentat.session import Session


# Make sure we always give posix paths to GPT
@pytest.mark.asyncio
async def test_posix_paths(mock_session_context):
    dir_name = "dir"
    file_name = "file.txt"
    file_path = Path(os.path.join(dir_name, file_name))
    os.makedirs(dir_name, exist_ok=True)
    with open(file_path, "w") as file_file:
        file_file.write("I am a file")
    mock_session_context.code_context.include(file_path)

    code_message = await mock_session_context.code_context.get_code_message(0)
    assert dir_name + "/" + file_name in code_message.split("\n")


@pytest.mark.asyncio
async def test_partial_files(mocker, mock_session_context):
    dir_name = "dir"
    file_name = "file.txt"
    file_path = os.path.join(dir_name, file_name)
    os.makedirs(dir_name, exist_ok=True)
    with open(file_path, "w") as file_file:
        file_file.write(dedent("""\
             I am a file
             with 5 lines
             third
             fourth
             fifth"""))

    file_path_partial = Path(file_path + ":1,3-5")
    mock_session_context.code_context.include(file_path_partial)
    mock_session_context.code_context.code_map = False

    code_message = await mock_session_context.code_context.get_code_message(0)
    assert code_message == dedent("""\
            Code Files:

            dir/file.txt
            1:I am a file
            ...
            3:third
            4:fourth
            """)


@pytest.mark.asyncio
async def test_run_from_subdirectory(
    temp_testbed,
    mock_collect_user_input,
    mock_call_llm_api,
):
    """Run mentat from a subdirectory of the git root"""
    # Change to the subdirectory
    os.chdir("multifile_calculator")
    mock_collect_user_input.set_stream_messages(
        [
            (
                "Insert the comment # Hello on the first line of"
                " multifile_calculator/calculator.py and scripts/echo.py"
            ),
            "y",
            "q",
        ]
    )
    mock_call_llm_api.set_streamed_values([dedent("""\
        I will insert a comment in both files.

        @@start
        {
            "file": "multifile_calculator/calculator.py",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }
        @@code
        # Hello
        @@end
        @@start
        {
            "file": "scripts/echo.py",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }
        @@code
        # Hello
        @@end""")])

    session = Session(
        cwd=temp_testbed, paths=["multifile_calculator/calculator.py", "scripts"]
    )
    session.start()
    await session.stream.recv(channel="client_exit")

    # Check that it works
    with open("calculator.py") as f:
        calculator_output = f.readlines()
    with open("../scripts/echo.py") as f:
        echo_output = f.readlines()
    assert calculator_output[0].strip() == "# Hello"
    assert echo_output[0].strip() == "# Hello"


@pytest.mark.asyncio
async def test_run_from_superdirectory(
    temp_testbed,
    mock_collect_user_input,
    mock_call_llm_api,
):
    """Run mentat from outside the git root"""
    # Change to the subdirectory
    mock_collect_user_input.set_stream_messages(
        [
            (
                "Insert the comment # Hello on the first line of"
                " multifile_calculator/calculator.py and scripts/echo.py"
            ),
            "y",
            "q",
        ]
    )
    mock_call_llm_api.set_streamed_values([dedent("""\
        I will insert a comment in both files.

        @@start
        {
            "file": "../multifile_calculator/calculator.py",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }
        @@code
        # Hello
        @@end
        @@start
        {
            "file": "../scripts/echo.py",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }
        @@code
        # Hello
        @@end""")])

    session = Session(
        cwd=Path(temp_testbed) / "format_examples",
        paths=["../multifile_calculator/calculator.py", "../scripts"],
    )
    session.start()
    await session.stream.recv(channel="client_exit")

    # Check that it works
    with open("multifile_calculator/calculator.py") as f:
        calculator_output = f.readlines()
    with open("scripts/echo.py") as f:
        echo_output = f.readlines()
    assert calculator_output[0].strip() == "# Hello"
    assert echo_output[0].strip() == "# Hello"


@pytest.mark.asyncio
async def test_change_after_creation(
    mock_collect_user_input,
    mock_call_llm_api,
):
    file_name = Path("hello_world.py")
    mock_collect_user_input.set_stream_messages(
        [
            "Conversation",
            "y",
            "q",
        ]
    )
    mock_call_llm_api.set_streamed_values([dedent(f"""\
        Conversation

        @@start
        {{
            "file": "{file_name}",
            "action": "create-file"
        }}
        @@code
        @@end
        @@start
        {{
            "file": "{file_name}",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }}
        @@code
        print("Hello, World!")
        @@end""")])

    session = Session(cwd=Path.cwd())
    session.start()
    await session.stream.recv(channel="client_exit")

    with file_name.open() as f:
        output = f.read()
    assert output == 'print("Hello, World!")'


@pytest.mark.asyncio
@pytest.mark.no_git_testbed
async def test_changed_file(
    mocker,
    temp_testbed,
    mock_collect_user_input,
    mock_session_context,
):
    dir_name = "dir"
    file_name = "file.txt"
    file_path = Path(temp_testbed) / dir_name / file_name
    os.makedirs(dir_name, exist_ok=True)
    with open(file_path, "w") as file_file:
        file_file.write("I am a file")

    code_context = mock_session_context.code_context
    code_context.include(file_path)

    # Load code_file_manager's file_lines
    code_file_manager = mock_session_context.code_file_manager
    code_file_manager.read_file(file_path)

    # Update the included files
    with open(file_path, "w") as file_file:
        file_file.write("I was a file")

    # Try to write_changes
    file_edit = FileEdit(
        file_path=file_path,
        replacements=[Replacement(0, 1, ["I am a file", "with edited lines"])],
    )
    assert file_edit.is_valid()

    # Decline overwrite
    mock_collect_user_input.set_stream_messages(["n", "q"])
    await code_file_manager.write_changes_to_files([file_edit])
    assert file_path.read_text().splitlines() == ["I was a file"]

    # Accept overwrite
    mock_collect_user_input.set_stream_messages(["y", "q"])
    await code_file_manager.write_changes_to_files([file_edit])
    assert file_path.read_text().splitlines() == ["I am a file", "with edited lines"]



# ruff: noqa: F401
# type: ignore

# Import all of the commands so that they are initialized
from .agent import AgentCommand
from .clear import ClearCommand
from .commit import CommitCommand
from .config import ConfigCommand
from .context import ContextCommand
from .exclude import ExcludeCommand
from .help import HelpCommand
from .include import IncludeCommand
from .redo import RedoCommand
from .run import RunCommand
from .sample import SampleCommand
from .screenshot import ScreenshotCommand
from .search import SearchCommand
from .talk import TalkCommand
from .undo import UndoCommand
from .undoall import UndoAllCommand
from .viewer import ViewerCommand



from pathlib import Path

from mentat.utils import get_relative_path


def test_get_relative_path(temp_testbed: Path):
    # Test path in same directory
    path = Path("__init__.py")
    target = temp_testbed
    assert get_relative_path(path, target) == Path("__init__.py")

    # Test path in parent directory
    path = Path("__init__.py")
    target = temp_testbed / "scripts"
    assert get_relative_path(path, target) == Path("../__init__.py")

    # Test path in child directory
    path = Path("multifile_calculator/__init__.py")
    target = temp_testbed
    assert get_relative_path(path, target) == Path("multifile_calculator/__init__.py")
